# Copyright 2014 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from testtools import matchers

from keystone.common import provider_api
from keystone.common import sql
from keystone.identity.mapping_backends import mapping
from keystone.tests import unit
from keystone.tests.unit import identity_mapping as mapping_sql
from keystone.tests.unit import test_backend_sql

PROVIDERS = provider_api.ProviderAPIs


class SqlIDMappingTable(test_backend_sql.SqlModels):
    """Set of tests for checking SQL Identity ID Mapping."""

    def test_id_mapping(self):
        cols = (
            ('public_id', sql.String, 64),
            ('domain_id', sql.String, 64),
            ('local_id', sql.String, 255),
            ('entity_type', sql.Enum, None),
        )
        self.assertExpectedSchema('id_mapping', cols)


class SqlIDMapping(test_backend_sql.SqlTests):
    def setUp(self):
        super().setUp()
        self.load_sample_data()

    def load_sample_data(self):
        self.addCleanup(self.clean_sample_data)
        domainA = unit.new_domain_ref()
        self.domainA = PROVIDERS.resource_api.create_domain(
            domainA['id'], domainA
        )
        domainB = unit.new_domain_ref()
        self.domainB = PROVIDERS.resource_api.create_domain(
            domainB['id'], domainB
        )

    def clean_sample_data(self):
        if hasattr(self, 'domainA'):
            self.domainA['enabled'] = False
            PROVIDERS.resource_api.update_domain(
                self.domainA['id'], self.domainA
            )
            PROVIDERS.resource_api.delete_domain(self.domainA['id'])
        if hasattr(self, 'domainB'):
            self.domainB['enabled'] = False
            PROVIDERS.resource_api.update_domain(
                self.domainB['id'], self.domainB
            )
            PROVIDERS.resource_api.delete_domain(self.domainB['id'])

    def test_invalid_public_key(self):
        self.assertIsNone(
            PROVIDERS.id_mapping_api.get_id_mapping(uuid.uuid4().hex)
        )

    def test_id_mapping_crud(self):
        initial_mappings = len(mapping_sql.list_id_mappings())
        local_id1 = uuid.uuid4().hex
        local_id2 = uuid.uuid4().hex
        local_entity1 = {
            'domain_id': self.domainA['id'],
            'local_id': local_id1,
            'entity_type': mapping.EntityType.USER,
        }
        local_entity2 = {
            'domain_id': self.domainB['id'],
            'local_id': local_id2,
            'entity_type': mapping.EntityType.GROUP,
        }

        # Check no mappings for the new local entities
        self.assertIsNone(
            PROVIDERS.id_mapping_api.get_public_id(local_entity1)
        )
        self.assertIsNone(
            PROVIDERS.id_mapping_api.get_public_id(local_entity2)
        )

        # Create the new mappings and then read them back
        public_id1 = PROVIDERS.id_mapping_api.create_id_mapping(local_entity1)
        public_id2 = PROVIDERS.id_mapping_api.create_id_mapping(local_entity2)
        self.assertThat(
            mapping_sql.list_id_mappings(),
            matchers.HasLength(initial_mappings + 2),
        )
        self.assertEqual(
            public_id1, PROVIDERS.id_mapping_api.get_public_id(local_entity1)
        )
        self.assertEqual(
            public_id2, PROVIDERS.id_mapping_api.get_public_id(local_entity2)
        )

        local_id_ref = PROVIDERS.id_mapping_api.get_id_mapping(public_id1)
        self.assertEqual(self.domainA['id'], local_id_ref['domain_id'])
        self.assertEqual(local_id1, local_id_ref['local_id'])
        self.assertEqual(mapping.EntityType.USER, local_id_ref['entity_type'])
        # Check we have really created a new external ID
        self.assertNotEqual(local_id1, public_id1)

        local_id_ref = PROVIDERS.id_mapping_api.get_id_mapping(public_id2)
        self.assertEqual(self.domainB['id'], local_id_ref['domain_id'])
        self.assertEqual(local_id2, local_id_ref['local_id'])
        self.assertEqual(mapping.EntityType.GROUP, local_id_ref['entity_type'])
        # Check we have really created a new external ID
        self.assertNotEqual(local_id2, public_id2)

        # Create another mappings, this time specifying a public ID to use
        new_public_id = uuid.uuid4().hex
        public_id3 = PROVIDERS.id_mapping_api.create_id_mapping(
            {
                'domain_id': self.domainB['id'],
                'local_id': local_id2,
                'entity_type': mapping.EntityType.USER,
            },
            public_id=new_public_id,
        )
        self.assertEqual(new_public_id, public_id3)
        self.assertThat(
            mapping_sql.list_id_mappings(),
            matchers.HasLength(initial_mappings + 3),
        )

        # Delete the mappings we created, and make sure the mapping count
        # goes back to where it was
        PROVIDERS.id_mapping_api.delete_id_mapping(public_id1)
        PROVIDERS.id_mapping_api.delete_id_mapping(public_id2)
        PROVIDERS.id_mapping_api.delete_id_mapping(public_id3)
        self.assertThat(
            mapping_sql.list_id_mappings(),
            matchers.HasLength(initial_mappings),
        )

    def test_id_mapping_handles_unicode(self):
        initial_mappings = len(mapping_sql.list_id_mappings())
        local_id = 'fäké1'
        local_entity = {
            'domain_id': self.domainA['id'],
            'local_id': local_id,
            'entity_type': mapping.EntityType.USER,
        }

        # Check no mappings for the new local entity
        self.assertIsNone(PROVIDERS.id_mapping_api.get_public_id(local_entity))

        # Create the new mapping and then read it back
        public_id = PROVIDERS.id_mapping_api.create_id_mapping(local_entity)
        self.assertThat(
            mapping_sql.list_id_mappings(),
            matchers.HasLength(initial_mappings + 1),
        )
        self.assertEqual(
            public_id, PROVIDERS.id_mapping_api.get_public_id(local_entity)
        )

    def test_id_mapping_handles_bytes(self):
        initial_mappings = len(mapping_sql.list_id_mappings())
        local_id = b'FaKeID'
        local_entity = {
            'domain_id': self.domainA['id'],
            'local_id': local_id,
            'entity_type': mapping.EntityType.USER,
        }

        # Check no mappings for the new local entity
        self.assertIsNone(PROVIDERS.id_mapping_api.get_public_id(local_entity))

        # Create the new mapping and then read it back
        public_id = PROVIDERS.id_mapping_api.create_id_mapping(local_entity)
        self.assertThat(
            mapping_sql.list_id_mappings(),
            matchers.HasLength(initial_mappings + 1),
        )
        self.assertEqual(
            public_id, PROVIDERS.id_mapping_api.get_public_id(local_entity)
        )

    def test_id_mapping_handles_ids_greater_than_64_characters(self):
        initial_mappings = len(mapping_sql.list_id_mappings())
        local_id = 'Aa' * 100
        local_entity = {
            'domain_id': self.domainA['id'],
            'local_id': local_id,
            'entity_type': mapping.EntityType.GROUP,
        }

        # Check no mappings for the new local entity
        self.assertIsNone(PROVIDERS.id_mapping_api.get_public_id(local_entity))

        # Create the new mapping and then read it back
        public_id = PROVIDERS.id_mapping_api.create_id_mapping(local_entity)
        self.assertThat(
            mapping_sql.list_id_mappings(),
            matchers.HasLength(initial_mappings + 1),
        )
        self.assertEqual(
            public_id, PROVIDERS.id_mapping_api.get_public_id(local_entity)
        )
        self.assertEqual(
            local_id,
            PROVIDERS.id_mapping_api.get_id_mapping(public_id)['local_id'],
        )

    def test_delete_public_id_is_silent(self):
        # Test that deleting an invalid public key is silent
        PROVIDERS.id_mapping_api.delete_id_mapping(uuid.uuid4().hex)

    def test_purge_mappings(self):
        initial_mappings = len(mapping_sql.list_id_mappings())
        local_id1 = uuid.uuid4().hex
        local_id2 = uuid.uuid4().hex
        local_id3 = uuid.uuid4().hex
        local_id4 = uuid.uuid4().hex
        local_id5 = uuid.uuid4().hex

        # Create five mappings,two in domainA, three in domainB
        PROVIDERS.id_mapping_api.create_id_mapping(
            {
                'domain_id': self.domainA['id'],
                'local_id': local_id1,
                'entity_type': mapping.EntityType.USER,
            }
        )
        PROVIDERS.id_mapping_api.create_id_mapping(
            {
                'domain_id': self.domainA['id'],
                'local_id': local_id2,
                'entity_type': mapping.EntityType.USER,
            }
        )
        public_id3 = PROVIDERS.id_mapping_api.create_id_mapping(
            {
                'domain_id': self.domainB['id'],
                'local_id': local_id3,
                'entity_type': mapping.EntityType.GROUP,
            }
        )
        public_id4 = PROVIDERS.id_mapping_api.create_id_mapping(
            {
                'domain_id': self.domainB['id'],
                'local_id': local_id4,
                'entity_type': mapping.EntityType.USER,
            }
        )
        public_id5 = PROVIDERS.id_mapping_api.create_id_mapping(
            {
                'domain_id': self.domainB['id'],
                'local_id': local_id5,
                'entity_type': mapping.EntityType.USER,
            }
        )

        self.assertThat(
            mapping_sql.list_id_mappings(),
            matchers.HasLength(initial_mappings + 5),
        )

        # Purge mappings for domainA, should be left with those in B
        PROVIDERS.id_mapping_api.purge_mappings(
            {'domain_id': self.domainA['id']}
        )
        self.assertThat(
            mapping_sql.list_id_mappings(),
            matchers.HasLength(initial_mappings + 3),
        )
        PROVIDERS.id_mapping_api.get_id_mapping(public_id3)
        PROVIDERS.id_mapping_api.get_id_mapping(public_id4)
        PROVIDERS.id_mapping_api.get_id_mapping(public_id5)

        # Purge mappings for type Group, should purge one more
        PROVIDERS.id_mapping_api.purge_mappings(
            {'entity_type': mapping.EntityType.GROUP}
        )
        self.assertThat(
            mapping_sql.list_id_mappings(),
            matchers.HasLength(initial_mappings + 2),
        )
        PROVIDERS.id_mapping_api.get_id_mapping(public_id4)
        PROVIDERS.id_mapping_api.get_id_mapping(public_id5)

        # Purge mapping for a specific local identifier
        PROVIDERS.id_mapping_api.purge_mappings(
            {
                'domain_id': self.domainB['id'],
                'local_id': local_id4,
                'entity_type': mapping.EntityType.USER,
            }
        )
        self.assertThat(
            mapping_sql.list_id_mappings(),
            matchers.HasLength(initial_mappings + 1),
        )
        PROVIDERS.id_mapping_api.get_id_mapping(public_id5)

        # Purge mappings the remaining mappings
        PROVIDERS.id_mapping_api.purge_mappings({})
        self.assertThat(
            mapping_sql.list_id_mappings(),
            matchers.HasLength(initial_mappings),
        )

    def test_create_duplicate_mapping(self):
        local_entity = {
            'domain_id': self.domainA['id'],
            'local_id': uuid.uuid4().hex,
            'entity_type': mapping.EntityType.USER,
        }
        public_id1 = PROVIDERS.id_mapping_api.create_id_mapping(local_entity)

        # second call should be successful and return the same
        # public_id as above
        public_id2 = PROVIDERS.id_mapping_api.create_id_mapping(local_entity)
        self.assertEqual(public_id1, public_id2)

        # even if public_id was specified, it should not be used,
        # and still the same public_id should be returned
        public_id3 = PROVIDERS.id_mapping_api.create_id_mapping(
            local_entity, public_id=uuid.uuid4().hex
        )
        self.assertEqual(public_id1, public_id3)

    @unit.skip_if_cache_disabled('identity')
    def test_cache_when_id_mapping_crud(self):
        local_id = uuid.uuid4().hex
        local_entity = {
            'domain_id': self.domainA['id'],
            'local_id': local_id,
            'entity_type': mapping.EntityType.USER,
        }

        # Check no mappings for the new local entity
        self.assertIsNone(PROVIDERS.id_mapping_api.get_public_id(local_entity))

        # Create new mappings, and it should be in the cache after created
        public_id = PROVIDERS.id_mapping_api.create_id_mapping(local_entity)
        self.assertEqual(
            public_id, PROVIDERS.id_mapping_api.get_public_id(local_entity)
        )
        local_id_ref = PROVIDERS.id_mapping_api.get_id_mapping(public_id)
        self.assertEqual(self.domainA['id'], local_id_ref['domain_id'])
        self.assertEqual(local_id, local_id_ref['local_id'])
        self.assertEqual(mapping.EntityType.USER, local_id_ref['entity_type'])

        # After delete the mapping, should be deleted from cache too
        PROVIDERS.id_mapping_api.delete_id_mapping(public_id)
        self.assertIsNone(PROVIDERS.id_mapping_api.get_public_id(local_entity))
        self.assertIsNone(PROVIDERS.id_mapping_api.get_id_mapping(public_id))

    @unit.skip_if_cache_disabled('identity')
    def test_invalidate_cache_when_purge_mappings(self):
        local_id1 = uuid.uuid4().hex
        local_id2 = uuid.uuid4().hex
        local_id3 = uuid.uuid4().hex
        local_id4 = uuid.uuid4().hex
        local_id5 = uuid.uuid4().hex

        # Create five mappings,two in domainA, three in domainB
        local_entity1 = {
            'domain_id': self.domainA['id'],
            'local_id': local_id1,
            'entity_type': mapping.EntityType.USER,
        }
        local_entity2 = {
            'domain_id': self.domainA['id'],
            'local_id': local_id2,
            'entity_type': mapping.EntityType.USER,
        }
        local_entity3 = {
            'domain_id': self.domainB['id'],
            'local_id': local_id3,
            'entity_type': mapping.EntityType.GROUP,
        }
        local_entity4 = {
            'domain_id': self.domainB['id'],
            'local_id': local_id4,
            'entity_type': mapping.EntityType.USER,
        }
        local_entity5 = {
            'domain_id': self.domainB['id'],
            'local_id': local_id5,
            'entity_type': mapping.EntityType.USER,
        }

        PROVIDERS.id_mapping_api.create_id_mapping(local_entity1)
        PROVIDERS.id_mapping_api.create_id_mapping(local_entity2)
        PROVIDERS.id_mapping_api.create_id_mapping(local_entity3)
        PROVIDERS.id_mapping_api.create_id_mapping(local_entity4)
        PROVIDERS.id_mapping_api.create_id_mapping(local_entity5)

        # Purge mappings for domainA, should be left with those in B
        PROVIDERS.id_mapping_api.purge_mappings(
            {'domain_id': self.domainA['id']}
        )
        self.assertIsNone(
            PROVIDERS.id_mapping_api.get_public_id(local_entity1)
        )
        self.assertIsNone(
            PROVIDERS.id_mapping_api.get_public_id(local_entity2)
        )

        # Purge mappings for type Group, should purge one more
        PROVIDERS.id_mapping_api.purge_mappings(
            {'entity_type': mapping.EntityType.GROUP}
        )
        self.assertIsNone(
            PROVIDERS.id_mapping_api.get_public_id(local_entity3)
        )

        # Purge mapping for a specific local identifier
        PROVIDERS.id_mapping_api.purge_mappings(
            {
                'domain_id': self.domainB['id'],
                'local_id': local_id4,
                'entity_type': mapping.EntityType.USER,
            }
        )
        self.assertIsNone(
            PROVIDERS.id_mapping_api.get_public_id(local_entity4)
        )

        # Purge mappings the remaining mappings
        PROVIDERS.id_mapping_api.purge_mappings({})
        self.assertIsNone(
            PROVIDERS.id_mapping_api.get_public_id(local_entity5)
        )

    def _prepare_domain_mappings_for_list(self):
        # Create five mappings:
        # two users in domainA, one group and two users in domainB
        local_entities = [
            {
                'domain_id': self.domainA['id'],
                'entity_type': mapping.EntityType.USER,
            },
            {
                'domain_id': self.domainA['id'],
                'entity_type': mapping.EntityType.USER,
            },
            {
                'domain_id': self.domainB['id'],
                'entity_type': mapping.EntityType.GROUP,
            },
            {
                'domain_id': self.domainB['id'],
                'entity_type': mapping.EntityType.USER,
            },
            {
                'domain_id': self.domainB['id'],
                'entity_type': mapping.EntityType.USER,
            },
        ]
        for e in local_entities:
            e['local_id'] = uuid.uuid4().hex
            e['public_id'] = PROVIDERS.id_mapping_api.create_id_mapping(e)
        return local_entities

    def test_get_domain_mapping_list(self):
        local_entities = self._prepare_domain_mappings_for_list()
        # NOTE(notmorgan): Always call to_dict in an active session context to
        # ensure that lazy-loaded relationships succeed. Edge cases could cause
        # issues especially in attribute mappers.
        with sql.session_for_read():
            # list mappings for domainA
            domain_a_mappings = (
                PROVIDERS.id_mapping_api.get_domain_mapping_list(
                    self.domainA['id']
                )
            )
            domain_a_mappings = [m.to_dict() for m in domain_a_mappings]
        self.assertCountEqual(local_entities[:2], domain_a_mappings)

    def test_get_domain_mapping_list_by_user_entity_type(self):
        local_entities = self._prepare_domain_mappings_for_list()
        # NOTE(notmorgan): Always call to_dict in an active session context to
        # ensure that lazy-loaded relationships succeed. Edge cases could cause
        # issues especially in attribute mappers.
        with sql.session_for_read():
            # list user mappings for domainB
            domain_b_mappings_user = (
                PROVIDERS.id_mapping_api.get_domain_mapping_list(
                    self.domainB['id'], entity_type=mapping.EntityType.USER
                )
            )
            domain_b_mappings_user = [
                m.to_dict() for m in domain_b_mappings_user
            ]
        self.assertCountEqual(local_entities[-2:], domain_b_mappings_user)

    def test_get_domain_mapping_list_by_group_entity_type(self):
        local_entities = self._prepare_domain_mappings_for_list()
        # NOTE(notmorgan): Always call to_dict in an active session context to
        # ensure that lazy-loaded relationships succeed. Edge cases could cause
        # issues especially in attribute mappers.
        with sql.session_for_read():
            # List group mappings for domainB. Given the data set, this should
            # only return a single reference, so don't both iterating the query
            # response.
            domain_b_mappings_group = (
                PROVIDERS.id_mapping_api.get_domain_mapping_list(
                    self.domainB['id'], entity_type=mapping.EntityType.GROUP
                )
            )
            domain_b_mappings_group = domain_b_mappings_group.first().to_dict()
        self.assertCountEqual(local_entities[2], domain_b_mappings_group)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import versionutils

from keystone.conf import utils

_DEPRECATED_MSG = utils.fmt(
    """
This option has been superseded by ephemeral users existing in the domain
of their identity provider.
"""
)

driver = cfg.StrOpt(
    'driver',
    default='sql',
    help=utils.fmt(
        """
Entry point for the federation backend driver in the `keystone.federation`
namespace. Keystone only provides a `sql` driver, so there is no reason to set
this option unless you are providing a custom entry point.
"""
    ),
)

assertion_prefix = cfg.StrOpt(
    'assertion_prefix',
    default='',
    help=utils.fmt(
        """
Prefix to use when filtering environment variable names for federated
assertions. Matched variables are passed into the federated mapping engine.
"""
    ),
)

remote_id_attribute = cfg.StrOpt(
    'remote_id_attribute',
    help=utils.fmt(
        """
Default value for all protocols to be used to obtain the entity ID of the
Identity Provider from the environment. For `mod_shib`, this would be
`Shib-Identity-Provider`. For `mod_auth_openidc`, this could be
`HTTP_OIDC_ISS`. For `mod_auth_mellon`, this could be `MELLON_IDP`. This can be
overridden on a per-protocol basis by providing a `remote_id_attribute` to the
federation protocol using the API.
"""
    ),
)

federated_domain_name = cfg.StrOpt(
    'federated_domain_name',
    default='Federated',
    deprecated_for_removal=True,
    deprecated_reason=_DEPRECATED_MSG,
    deprecated_since=versionutils.deprecated.TRAIN,
    help=utils.fmt(
        """
An arbitrary domain name that is reserved to allow federated ephemeral users to
have a domain concept. Note that an admin will not be able to create a domain
with this name or update an existing domain to this name. You are not advised
to change this value unless you really have to.
"""
    ),
)

trusted_dashboard = cfg.MultiStrOpt(
    'trusted_dashboard',
    default=[],
    help=utils.fmt(
        """
A list of trusted dashboard hosts. Before accepting a Single Sign-On request to
return a token, the origin host must be a member of this list. This
configuration option may be repeated for multiple values. You must set this in
order to use web-based SSO flows. For example:
trusted_dashboard=https://acme.example.com/auth/websso
trusted_dashboard=https://beta.example.com/auth/websso
"""
    ),
)

sso_callback_template = cfg.StrOpt(
    'sso_callback_template',
    default='/etc/keystone/sso_callback_template.html',
    help=utils.fmt(
        """
Absolute path to an HTML file used as a Single Sign-On callback handler. This
page is expected to redirect the user from keystone back to a trusted dashboard
host, by form encoding a token in a POST request. Keystone's default value
should be sufficient for most deployments.
"""
    ),
)


caching = cfg.BoolOpt(
    'caching',
    default=True,
    help=utils.fmt(
        """
Toggle for federation caching. This has no effect unless global caching is
enabled. There is typically no reason to disable this.
"""
    ),
)


default_authorization_ttl = cfg.IntOpt(
    'default_authorization_ttl',
    default=0,
    help=utils.fmt(
        """
Default time in minutes for the validity of group memberships carried over
from a mapping. Default is 0, which means disabled.
"""
    ),
)

attribute_mapping_default_schema_version = cfg.StrOpt(
    'attribute_mapping_default_schema_version',
    default='1.0',
    help=utils.fmt(
        """
The attribute mapping default schema version to be used, if the attribute
mapping being registered does not have a schema version. One must bear in
mind that changing this value will have no effect on attribute mappings that
were previously registered when another default value was applied. Once
registered, one needs to update the attribute mapping schema via the update
API to be able to change an attribute mapping schema version.
"""
    ),
)


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [
    driver,
    assertion_prefix,
    remote_id_attribute,
    federated_domain_name,
    trusted_dashboard,
    sso_callback_template,
    caching,
    default_authorization_ttl,
    attribute_mapping_default_schema_version,
]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Request body validating middleware for OpenStack Identity resources."""

from keystone.common.validation import validators


def lazy_validate(request_body_schema, resource_to_validate):
    """A non-decorator way to validate a request, to be used inline.

    :param request_body_schema: a schema to validate the resource reference
    :param resource_to_validate: dictionary to validate
    :raises keystone.exception.ValidationError: if `resource_to_validate` is
            None. (see wrapper method below).
    :raises TypeError: at decoration time when the expected resource to
                       validate isn't found in the decorated method's
                       signature

    """
    schema_validator = validators.SchemaValidator(request_body_schema)
    schema_validator.validate(resource_to_validate)


def nullable(property_schema):
    """Clone a property schema into one that is nullable.

    :param dict property_schema: schema to clone into a nullable schema
    :returns: a new dict schema
    """
    # TODO(dstanek): deal with the case where type is already a list; we don't
    #                do that yet so I'm not wasting time on it
    new_schema = property_schema.copy()
    new_schema['type'] = [property_schema['type'], 'null']
    # NOTE(kmalloc): If enum is specified (such as our boolean case) ensure we
    # add null to the enum as well so that null can be passed/validated as
    # expected. Without adding to the enum, null will not validate as enum is
    # explicitly listing valid values. According to the JSON Schema
    # specification, the values must be unique in the enum array.
    if 'enum' in new_schema and None not in new_schema['enum']:
        # In the enum the 'null' is NoneType
        new_schema['enum'].append(None)
    return new_schema



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import uuid

from keystone.common import driver_hints
from keystone.common import provider_api

PROVIDERS = provider_api.ProviderAPIs


class ShadowUsersCoreTests:
    def test_shadow_federated_user(self):
        federated_user1 = copy.deepcopy(self.federated_user)
        ShadowUsersCoreTests.normalize_federated_user_properties_for_test(
            federated_user1, email=self.email
        )

        user = PROVIDERS.identity_api.shadow_federated_user(
            self.federated_user['idp_id'],
            self.federated_user['protocol_id'],
            federated_user1,
        )

        self.assertIsNotNone(user['id'])
        self.assertEqual(7, len(user.keys()))
        self.assertIsNotNone(user['name'])
        self.assertIsNone(user['password_expires_at'])
        self.assertIsNotNone(user['domain_id'])

        # NOTE(breton): below, attribute `enabled` is explicitly tested to be
        # equal True. assertTrue should not be used, because it converts
        # the passed value to bool().
        self.assertEqual(True, user['enabled'])
        self.assertIsNotNone(user['email'])

    def test_shadow_existing_federated_user(self):
        federated_user1 = copy.deepcopy(self.federated_user)
        ShadowUsersCoreTests.normalize_federated_user_properties_for_test(
            federated_user1, email=self.email
        )

        # introduce the user to keystone for the first time
        shadow_user1 = PROVIDERS.identity_api.shadow_federated_user(
            self.federated_user['idp_id'],
            self.federated_user['protocol_id'],
            federated_user1,
        )

        self.assertEqual(federated_user1['display_name'], shadow_user1['name'])

        # shadow the user again, with another name to invalidate the cache
        # internally, this operation causes request to the driver. It should
        # not fail.
        federated_user2 = copy.deepcopy(self.federated_user)
        federated_user2['display_name'] = uuid.uuid4().hex
        ShadowUsersCoreTests.normalize_federated_user_properties_for_test(
            federated_user2, email=self.email
        )

        shadow_user2 = PROVIDERS.identity_api.shadow_federated_user(
            self.federated_user['idp_id'],
            self.federated_user['protocol_id'],
            federated_user2,
        )
        self.assertEqual(federated_user2['display_name'], shadow_user2['name'])
        self.assertNotEqual(shadow_user1['name'], shadow_user2['name'])

        # The shadowed users still share the same unique ID.
        self.assertEqual(shadow_user1['id'], shadow_user2['id'])

    def test_shadow_federated_user_not_creating_a_local_user(self):
        federated_user1 = copy.deepcopy(self.federated_user)
        ShadowUsersCoreTests.normalize_federated_user_properties_for_test(
            federated_user1, email="some_id@mail.provider"
        )

        PROVIDERS.identity_api.shadow_federated_user(
            federated_user1['idp_id'],
            federated_user1['protocol_id'],
            federated_user1,
        )

        hints = driver_hints.Hints()
        hints.add_filter('name', federated_user1['display_name'])
        users = PROVIDERS.identity_api.list_users(hints=hints)

        self.assertEqual(1, len(users))

        federated_user2 = copy.deepcopy(federated_user1)
        # Avoid caching
        federated_user2['name'] = uuid.uuid4().hex
        federated_user2['id'] = uuid.uuid4().hex
        federated_user2['email'] = "some_id_2@mail.provider"

        PROVIDERS.identity_api.shadow_federated_user(
            federated_user2['idp_id'],
            federated_user2['protocol_id'],
            federated_user2,
        )

        hints.add_filter('name', federated_user2['display_name'])
        users = PROVIDERS.identity_api.list_users(hints=hints)

        # The number os users must remain 1
        self.assertEqual(1, len(users))

    @staticmethod
    def normalize_federated_user_properties_for_test(
        federated_user, email=None
    ):
        federated_user['email'] = email
        federated_user['id'] = federated_user['unique_id']
        federated_user['name'] = federated_user['display_name']
        if not federated_user.get('domain'):
            federated_user['domain'] = {'id': uuid.uuid4().hex}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from sqlalchemy import orm
from sqlalchemy.orm import collections

from keystone.assignment.role_backends import base
from keystone.assignment.role_backends import resource_options as ro
from keystone.common import resource_options
from keystone.common import sql


class RoleTable(sql.ModelBase, sql.ModelDictMixinWithExtras):
    def to_dict(self, include_extra_dict=False):
        d = super().to_dict(include_extra_dict=include_extra_dict)
        if d['domain_id'] == base.NULL_DOMAIN_ID:
            d['domain_id'] = None
        # NOTE(notmorgan): Eventually it may make sense to drop the empty
        # option dict creation to the superclass (if enough models use it)
        d['options'] = resource_options.ref_mapper_to_dict_options(self)
        return d

    @classmethod
    def from_dict(cls, role_dict):
        if 'domain_id' in role_dict and role_dict['domain_id'] is None:
            new_dict = role_dict.copy()
            new_dict['domain_id'] = base.NULL_DOMAIN_ID
        else:
            new_dict = role_dict
        # TODO(morgan): move this functionality to a common location
        resource_options = {}
        options = new_dict.pop('options', {})
        for opt in cls.resource_options_registry.options:
            if opt.option_name in options:
                opt_value = options[opt.option_name]
                # NOTE(notmorgan): None is always a valid type
                if opt_value is not None:
                    opt.validator(opt_value)
                resource_options[opt.option_id] = opt_value
        role_obj = super().from_dict(new_dict)
        setattr(role_obj, '_resource_options', resource_options)
        return role_obj

    __tablename__ = 'role'
    attributes = ['id', 'name', 'domain_id', 'description']
    resource_options_registry = ro.ROLE_OPTIONS_REGISTRY
    id = sql.Column(sql.String(64), primary_key=True)
    name = sql.Column(sql.String(255), nullable=False)
    domain_id = sql.Column(
        sql.String(64), nullable=False, server_default=base.NULL_DOMAIN_ID
    )
    description = sql.Column(sql.String(255), nullable=True)
    extra = sql.Column(sql.JsonBlob())
    _resource_option_mapper = orm.relationship(
        'RoleOption',
        single_parent=True,
        cascade='all,delete,delete-orphan',
        lazy='subquery',
        backref='role',
        collection_class=collections.attribute_mapped_collection('option_id'),
    )
    __table_args__ = (sql.UniqueConstraint('name', 'domain_id'),)


class ImpliedRoleTable(sql.ModelBase, sql.ModelDictMixin):
    __tablename__ = 'implied_role'
    attributes = ['prior_role_id', 'implied_role_id']
    prior_role_id = sql.Column(
        sql.String(64),
        sql.ForeignKey('role.id', ondelete="CASCADE"),
        primary_key=True,
    )
    implied_role_id = sql.Column(
        sql.String(64),
        sql.ForeignKey('role.id', ondelete="CASCADE"),
        primary_key=True,
    )

    @classmethod
    def from_dict(cls, dictionary):
        new_dictionary = dictionary.copy()
        return cls(**new_dictionary)

    def to_dict(self):
        """Return a dictionary with model's attributes.

        overrides the `to_dict` function from the base class
        to avoid having an `extra` field.
        """
        d = {}
        for attr in self.__class__.attributes:
            d[attr] = getattr(self, attr)
        return d


class RoleOption(sql.ModelBase):
    __tablename__ = 'role_option'
    role_id = sql.Column(
        sql.String(64),
        sql.ForeignKey('role.id', ondelete='CASCADE'),
        nullable=False,
        primary_key=True,
    )
    option_id = sql.Column(sql.String(4), nullable=False, primary_key=True)
    option_value = sql.Column(sql.JsonBlob, nullable=True)

    def __init__(self, option_id, option_value):
        self.option_id = option_id
        self.option_value = option_value



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import typing as ty

from alembic import command as alembic_api
from alembic import script as alembic_script
import fixtures
from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import test_fixtures
from oslo_log import log as logging
import sqlalchemy

# We need to import all of these so the tables are registered. It would be
# easier if these were all in a central location :(
import keystone.application_credential.backends.sql  # noqa: F401
import keystone.assignment.backends.sql  # noqa: F401
import keystone.assignment.role_backends.sql_model  # noqa: F401
import keystone.catalog.backends.sql  # noqa: F401
from keystone.common import sql
from keystone.common.sql import upgrades
import keystone.conf
import keystone.credential.backends.sql  # noqa: F401
import keystone.endpoint_policy.backends.sql  # noqa: F401
import keystone.federation.backends.sql  # noqa: F401
import keystone.identity.backends.sql_model  # noqa: F401
import keystone.identity.mapping_backends.sql  # noqa: F401
import keystone.limit.backends.sql  # noqa: F401
import keystone.oauth1.backends.sql  # noqa: F401
import keystone.policy.backends.sql  # noqa: F401
import keystone.resource.backends.sql_model  # noqa: F401
import keystone.resource.config_backends.sql  # noqa: F401
import keystone.revoke.backends.sql  # noqa: F401
from keystone.tests import unit
import keystone.trust.backends.sql  # noqa: F401

LOG = logging.getLogger(__name__)


class DBOperationNotAllowed(Exception):
    pass


class BannedDBSchemaOperations(fixtures.Fixture):
    """Ban some operations for migrations."""

    def __init__(self, banned_ops, revision):
        super().__init__()
        self._banned_ops = banned_ops or {}
        self._revision = revision

    @staticmethod
    def _explode(op, revision):
        def fail(*a, **kw):
            msg = "Operation '%s' is not allowed in migration %s"
            raise DBOperationNotAllowed(msg % (op, revision))

        return fail

    def setUp(self):
        super().setUp()
        for op in self._banned_ops:
            self.useFixture(
                fixtures.MonkeyPatch(
                    f'alembic.op.{op}', self._explode(op, self._revision)
                )
            )


class KeystoneMigrationsWalk(test_fixtures.OpportunisticDBTestMixin):
    # Migrations can take a long time, particularly on underpowered CI nodes.
    # Give them some breathing room.
    TIMEOUT_SCALING_FACTOR = 4

    BANNED_OPS = {
        'expand': [
            'alter_column',
            'drop_column',
            'drop_constraint',
            'drop_index',
            'drop_table',
            'drop_table_comment',
            # 'execute',
            'rename_table',
        ],
        'contract': {
            'add_column',
            'bulk_insert',
            'create_check_constraint',
            'create_exclude_constraint',
            'create_foreign_key',
            'create_index',
            'create_primary_key',
            'create_table',
            'create_table_comment',
            'create_unique_constraint',
            # 'execute',
            'rename_table',
        },
    }

    BANNED_OP_EXCEPTIONS: list[ty.Any] = [
        # NOTE(xek, henry-nash): Reviewers: DO NOT ALLOW THINGS TO BE ADDED
        # HERE UNLESS JUSTIFICATION CAN BE PROVIDED AS TO WHY THIS WILL NOT
        # CAUSE PROBLEMS FOR ROLLING UPGRADES.
    ]

    def setUp(self):
        super().setUp()
        self.engine = enginefacade.writer.get_engine()
        self.config = upgrades._find_alembic_conf()
        self.init_version = upgrades.ALEMBIC_INIT_VERSION

        # TODO(stephenfin): Do we need this? I suspect not since we're using
        # enginefacade.write.get_engine() directly above
        # Override keystone's context manager to be oslo.db's global context
        # manager.
        sql.core._TESTING_USE_GLOBAL_CONTEXT_MANAGER = True
        self.addCleanup(
            setattr, sql.core, '_TESTING_USE_GLOBAL_CONTEXT_MANAGER', False
        )
        self.addCleanup(sql.cleanup)

    def _migrate_up(self, connection, revision):
        version = revision.revision

        if version == self.init_version:  # no tests for the initial revision
            alembic_api.upgrade(self.config, version)
            return

        self.assertIsNotNone(
            getattr(self, f'_check_{version}', None),
            (f'DB Migration {version} does not have a test; you must add one'),
        )

        pre_upgrade = getattr(self, f'_pre_upgrade_{version}', None)
        if pre_upgrade:
            pre_upgrade(connection)

        banned_ops = []
        if version not in self.BANNED_OP_EXCEPTIONS:
            # there should only ever be one label, but this is safer
            for branch_label in revision.branch_labels:
                banned_ops.extend(self.BANNED_OPS[branch_label])

        # SQLite migrations are running in batch mode, which mean we recreate a
        # table in all migrations. As such, we can't really blacklist things so
        # don't even try.
        if self.FIXTURE.DRIVER == 'sqlite':
            banned_ops = []

        with BannedDBSchemaOperations(banned_ops, version):
            alembic_api.upgrade(self.config, version)

        post_upgrade = getattr(self, f'_check_{version}', None)
        if post_upgrade:
            post_upgrade(connection)

    def _pre_upgrade_e25ffa003242(self, connection):
        """This is a no-op migration."""
        pass

    def _check_e25ffa003242(self, connection):
        """This is a no-op migration."""
        pass

    def _pre_upgrade_29e87d24a316(self, connection):
        """This is a no-op migration."""
        pass

    def _check_29e87d24a316(self, connection):
        """This is a no-op migration."""
        pass

    # 2023.2 bobcat

    _99de3849d860_removed_constraints = {
        'access_rule': 'access_rule_external_id_key',
        'trust': 'duplicate_trust_constraint_expanded',
    }

    def _pre_upgrade_99de3849d860(self, connection):
        inspector = sqlalchemy.inspect(connection)
        for (
            table,
            constraint,
        ) in self._99de3849d860_removed_constraints.items():
            constraints = [
                x['name'] for x in inspector.get_unique_constraints(table)
            ]
            self.assertIn(constraint, constraints)

    def _check_99de3849d860(self, connection):
        inspector = sqlalchemy.inspect(connection)
        for (
            table,
            constraint,
        ) in self._99de3849d860_removed_constraints.items():
            constraints = [
                x['name'] for x in inspector.get_unique_constraints(table)
            ]
            self.assertNotIn(constraint, constraints)

    def _pre_upgrade_b4f8b3f584e0(self, connection):
        inspector = sqlalchemy.inspect(connection)
        constraints = inspector.get_unique_constraints('trust')
        self.assertNotIn(
            'duplicate_trust_constraint', {x['name'] for x in constraints}
        )

        all_constraints = []
        for c in constraints:
            all_constraints + c.get('column_names', [])

        not_allowed_constraints = [
            'trustor_user_id',
            'trustee_user_id',
            'project_id',
            'impersonation',
            'expires_at',
        ]
        for not_c in not_allowed_constraints:
            self.assertNotIn(not_c, all_constraints)

    def _check_b4f8b3f584e0(self, connection):
        inspector = sqlalchemy.inspect(connection)
        constraints = inspector.get_unique_constraints('trust')
        self.assertIn(
            'duplicate_trust_constraint', {x['name'] for x in constraints}
        )
        constraint = [
            x for x in constraints if x['name'] == 'duplicate_trust_constraint'
        ][0]
        self.assertEqual(
            [
                'trustor_user_id',
                'trustee_user_id',
                'project_id',
                'impersonation',
                'expires_at',
            ],
            constraint['column_names'],
        )

    def _pre_upgrade_c88cdce8f248(self, connection):
        if connection.engine.name != 'mysql':
            return

        # NOTE(stephenfin): Even though the migration is written to handle
        # names generated by both alembic and sqlalchemy-migrate, we only check
        # for the former here since we don't apply sqlalchemy-migrate
        # migrations anymore
        inspector = sqlalchemy.inspect(connection)
        indexes = inspector.get_indexes('project_tag')
        self.assertIn('project_id', {x['name'] for x in indexes})

    def _check_c88cdce8f248(self, connection):
        # This migration only applies to MySQL
        if connection.engine.name != 'mysql':
            return

        inspector = sqlalchemy.inspect(connection)
        indexes = inspector.get_indexes('project_tag')
        self.assertNotIn('project_id', {x['name'] for x in indexes})

    def _pre_upgrade_11c3b243b4cb(self, connection):
        inspector = sqlalchemy.inspect(connection)
        columns = inspector.get_columns('service_provider')
        found = False
        for column in columns:
            if column['name'] != 'relay_state_prefix':
                continue

            # The default should initially be set to the CONF value
            self.assertIsNotNone(column['default'])
            found = True
        self.assertTrue(found, 'Failed to find column')

    def _check_11c3b243b4cb(self, connection):
        inspector = sqlalchemy.inspect(connection)
        columns = inspector.get_columns('service_provider')
        found = False
        for column in columns:
            if column['name'] != 'relay_state_prefix':
                continue

            # The default should now be unset
            self.assertIsNone(column['default'])
            found = True
        self.assertTrue(found, 'Failed to find column')

    def _pre_upgrade_47147121(self, connection):
        inspector = sqlalchemy.inspect(connection)
        columns = inspector.get_columns('mapping')

        all_column_names = []
        for c in columns:
            all_column_names.append(c.get('name'))

        self.assertNotIn('schema_version', all_column_names)

    def _check_47147121(self, connection):
        inspector = sqlalchemy.inspect(connection)
        columns = inspector.get_columns('mapping')

        all_column_names = []
        for c in columns:
            all_column_names.append(c.get('name'))

        self.assertIn('schema_version', all_column_names)

    def test_single_base_revision(self):
        """Ensure we only have a single base revision.

        There's no good reason for us to have diverging history, so validate
        that only one base revision exists. This will prevent simple errors
        where people forget to specify the base revision. If this fail for
        your change, look for migrations that do not have a 'revises' line in
        them.
        """
        script = alembic_script.ScriptDirectory.from_config(self.config)
        self.assertEqual(1, len(script.get_bases()))

    def test_head_revisions(self):
        """Ensure we only have a two head revisions.

        There's no good reason for us to have diverging history beyond the
        expand and contract branches, so validate that only these head
        revisions exist. This will prevent merge conflicts adding additional
        head revision points. If this fail for your change, look for migrations
        with the duplicate 'revises' line in them.
        """
        script = alembic_script.ScriptDirectory.from_config(self.config)
        self.assertEqual(2, len(script.get_heads()))

    def test_walk_versions(self):
        with self.engine.begin() as connection:
            self.config.attributes['connection'] = connection
            script = alembic_script.ScriptDirectory.from_config(self.config)
            revisions = list(script.walk_revisions())

            # for some reason, 'walk_revisions' gives us the revisions in
            # reverse chronological order so we have to invert this
            revisions.reverse()
            self.assertEqual(revisions[0].revision, self.init_version)

            for revision in revisions:
                LOG.info('Testing revision %s', revision.revision)
                self._migrate_up(connection, revision)

    def _get_head_from_file(self, branch):
        path = os.path.join(
            os.path.dirname(upgrades.__file__),
            'migrations',
            'versions',
            f'{branch.upper()}_HEAD',
        )

        with open(path) as fh:
            return fh.read().strip()

    def test_db_version_alembic(self):
        upgrades.offline_sync_database_to_version(engine=self.engine)

        for branch in (upgrades.EXPAND_BRANCH, upgrades.CONTRACT_BRANCH):
            head = self._get_head_from_file(branch)
            self.assertEqual(head, upgrades.get_db_version(branch))


class TestMigrationsWalkSQLite(
    KeystoneMigrationsWalk,
    test_fixtures.OpportunisticDBTestMixin,
    unit.TestCase,
):
    pass


class TestMigrationsWalkMySQL(
    KeystoneMigrationsWalk,
    test_fixtures.OpportunisticDBTestMixin,
    unit.TestCase,
):
    FIXTURE = test_fixtures.MySQLOpportunisticFixture


class TestMigrationsWalkPostgreSQL(
    KeystoneMigrationsWalk,
    test_fixtures.OpportunisticDBTestMixin,
    unit.TestCase,
):
    FIXTURE = test_fixtures.PostgresqlOpportunisticFixture



# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Token provider interface."""

import base64
import datetime
import uuid

from oslo_log import log
from oslo_utils import timeutils

from keystone.common import cache
from keystone.common import manager
from keystone.common import provider_api
from keystone.common import utils
import keystone.conf
from keystone import exception
from keystone.federation import constants
from keystone.i18n import _
from keystone.models import token_model
from keystone import notifications

CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)
PROVIDERS = provider_api.ProviderAPIs

TOKENS_REGION = cache.create_region(name='tokens')
MEMOIZE_TOKENS = cache.get_memoization_decorator(
    group='token', region=TOKENS_REGION
)

# NOTE(morganfainberg): This is for compatibility in case someone was relying
# on the old location of the UnsupportedTokenVersionException for their code.
UnsupportedTokenVersionException = exception.UnsupportedTokenVersionException

# supported token versions
V3 = token_model.V3
VERSIONS = token_model.VERSIONS

# minimum access rules support
ACCESS_RULES_MIN_VERSION = token_model.ACCESS_RULES_MIN_VERSION


def default_expire_time():
    """Determine when a fresh token should expire.

    Expiration time varies based on configuration (see ``[token] expiration``).

    :returns: a naive UTC datetime.datetime object

    """
    expire_delta = datetime.timedelta(seconds=CONF.token.expiration)
    expires_at = timeutils.utcnow() + expire_delta
    return expires_at.replace(microsecond=0)


def random_urlsafe_str():
    """Generate a random URL-safe string.

    :rtype: str
    """
    # chop the padding (==) off the end of the encoding to save space
    return base64.urlsafe_b64encode(uuid.uuid4().bytes)[:-2].decode('utf-8')


class Manager(manager.Manager):
    """Default pivot point for the token provider backend.

    See :mod:`keystone.common.manager.Manager` for more details on how this
    dynamically calls the backend.

    """

    driver_namespace = 'keystone.token.provider'
    _provides_api = 'token_provider_api'

    V3 = V3
    VERSIONS = VERSIONS

    def __init__(self):
        super().__init__(CONF.token.provider)
        self._register_callback_listeners()

    def _register_callback_listeners(self):
        # This is used by the @dependency.provider decorator to register the
        # provider (token_provider_api) manager to listen for trust deletions.
        callbacks = {
            notifications.ACTIONS.deleted: [
                ['OS-TRUST:trust', self._drop_token_cache],
                ['user', self._drop_token_cache],
                ['domain', self._drop_token_cache],
            ],
            notifications.ACTIONS.disabled: [
                ['user', self._drop_token_cache],
                ['domain', self._drop_token_cache],
                ['project', self._drop_token_cache],
            ],
            notifications.ACTIONS.internal: [
                [notifications.INVALIDATE_TOKEN_CACHE, self._drop_token_cache]
            ],
        }

        for event, cb_info in callbacks.items():
            for resource_type, callback_fns in cb_info:
                notifications.register_event_callback(
                    event, resource_type, callback_fns
                )

    def _drop_token_cache(self, service, resource_type, operation, payload):
        """Invalidate the entire token cache.

        This is a handy private utility method that should be used when
        consuming notifications that signal invalidating the token cache.

        """
        if CONF.token.cache_on_issue or CONF.token.caching:
            TOKENS_REGION.invalidate()

    @MEMOIZE_TOKENS
    def check_revocation_v3(self, token_values):
        PROVIDERS.revoke_api.check_token(token_values)

    def check_revocation(self, token):
        token_values = self.revoke_api.model.build_token_values(token)
        return self.check_revocation_v3(token_values)

    def validate_token(
        self, token_id, window_seconds=0, access_rules_support=None
    ):
        if not token_id:
            raise exception.TokenNotFound(_('No token in the request'))

        try:
            token = self._validate_token(token_id)
            self._is_valid_token(token, window_seconds=window_seconds)
            self._validate_token_access_rules(token, access_rules_support)
            return token
        except exception.Unauthorized as e:
            LOG.debug('Unable to validate token: %s', e)
            raise exception.TokenNotFound(token_id=token_id)

    @MEMOIZE_TOKENS
    def _validate_token(self, token_id):
        (
            user_id,
            methods,
            audit_ids,
            system,
            domain_id,
            project_id,
            trust_id,
            federated_group_ids,
            identity_provider_id,
            protocol_id,
            access_token_id,
            app_cred_id,
            thumbprint,
            issued_at,
            expires_at,
        ) = self.driver.validate_token(token_id)

        token = token_model.TokenModel()
        token.user_id = user_id
        token.methods = methods
        if len(audit_ids) > 1:
            token.parent_audit_id = audit_ids.pop()
        token.audit_id = audit_ids.pop()
        token.system = system
        token.domain_id = domain_id
        token.project_id = project_id
        token.trust_id = trust_id
        token.access_token_id = access_token_id
        token.application_credential_id = app_cred_id
        token.oauth2_thumbprint = thumbprint
        token.expires_at = expires_at
        if federated_group_ids is not None:
            token.is_federated = True
            token.identity_provider_id = identity_provider_id
            token.protocol_id = protocol_id
            token.federated_groups = federated_group_ids

        token.mint(token_id, issued_at)
        return token

    def _is_valid_token(self, token, window_seconds=0):
        """Verify the token is valid format and has not expired."""
        current_time = timeutils.normalize_time(timeutils.utcnow())

        try:
            expiry = timeutils.parse_isotime(token.expires_at)
            expiry = timeutils.normalize_time(expiry)

            # add a window in which you can fetch a token beyond expiry
            expiry += datetime.timedelta(seconds=window_seconds)

        except Exception:
            LOG.exception(
                'Unexpected error or malformed token '
                'determining token expiry: %s',
                token,
            )
            raise exception.TokenNotFound(_('Failed to validate token'))

        if current_time < expiry:
            self.check_revocation(token)
            # Token has not expired and has not been revoked.
            return None
        else:
            raise exception.TokenNotFound(_('Failed to validate token'))

    def _validate_token_access_rules(self, token, access_rules_support=None):
        if token.application_credential_id:
            app_cred_api = PROVIDERS.application_credential_api
            app_cred = app_cred_api.get_application_credential(
                token.application_credential_id
            )
            if app_cred.get('access_rules') is not None and (
                not access_rules_support
                or (float(access_rules_support) < ACCESS_RULES_MIN_VERSION)
            ):
                LOG.exception(
                    'Attempted to use application credential'
                    ' access rules with a middleware that does not'
                    ' understand them. You must upgrade'
                    ' keystonemiddleware on all services that'
                    ' accept application credentials as an'
                    ' authentication method.'
                )
                raise exception.TokenNotFound(_('Failed to validate token'))

    def issue_token(
        self,
        user_id,
        method_names,
        expires_at=None,
        system=None,
        project_id=None,
        domain_id=None,
        auth_context=None,
        trust_id=None,
        app_cred_id=None,
        thumbprint=None,
        parent_audit_id=None,
    ):
        # NOTE(lbragstad): Grab a blank token object and use composition to
        # build the token according to the authentication and authorization
        # context. This cuts down on the amount of logic we have to stuff into
        # the TokenModel's __init__() method.
        token = token_model.TokenModel()
        token.methods = method_names
        token.system = system
        token.domain_id = domain_id
        token.project_id = project_id
        token.trust_id = trust_id
        token.application_credential_id = app_cred_id
        token.audit_id = random_urlsafe_str()
        token.oauth2_thumbprint = thumbprint
        token.parent_audit_id = parent_audit_id

        if auth_context:
            if constants.IDENTITY_PROVIDER in auth_context:
                token.is_federated = True
                token.protocol_id = auth_context[constants.PROTOCOL]
                idp_id = auth_context[constants.IDENTITY_PROVIDER]
                if isinstance(idp_id, bytes):
                    idp_id = idp_id.decode('utf-8')
                token.identity_provider_id = idp_id
                token.user_id = auth_context['user_id']
                token.federated_groups = [
                    {'id': group} for group in auth_context['group_ids']
                ]

            if 'access_token_id' in auth_context:
                token.access_token_id = auth_context['access_token_id']

        if not token.user_id:
            token.user_id = user_id

        token.user_domain_id = token.user['domain_id']

        if isinstance(expires_at, datetime.datetime):
            token.expires_at = utils.isotime(expires_at, subsecond=True)
        if isinstance(expires_at, str):
            token.expires_at = expires_at
        elif not expires_at:
            token.expires_at = utils.isotime(
                default_expire_time(), subsecond=True
            )

        # NOTE(d34dh0r53): If this token is being issued with an application
        # credential and the application credential expires before the token
        # we need to set the token expiration to be the same as the application
        # credential. See CVE-2022-2447 for more information.
        if app_cred_id is not None:
            app_cred_api = PROVIDERS.application_credential_api
            app_cred = app_cred_api.get_application_credential(
                token.application_credential_id
            )
            token_time = timeutils.normalize_time(
                timeutils.parse_isotime(token.expires_at)
            )
            if (app_cred['expires_at'] is not None) and (
                token_time > app_cred['expires_at']
            ):
                token.expires_at = utils.isotime(
                    app_cred['expires_at'], subsecond=True
                )
                LOG.debug(
                    'Resetting token expiration to the application'
                    ' credential expiration: %s',
                    token.expires_at,
                )

        token_id, issued_at = self.driver.generate_id_and_issued_at(token)
        token.mint(token_id, issued_at)

        # cache the token object and with ID
        if CONF.token.cache_on_issue or CONF.token.caching:
            # NOTE(amakarov): here and above TOKENS_REGION is to be passed
            # to serve as required positional "self" argument. It's ignored,
            # so I've put it here for convenience - any placeholder is fine.
            self._validate_token.set(token, self, token.id)

        return token

    def invalidate_individual_token_cache(self, token):
        # NOTE(morganfainberg): invalidate takes the exact same arguments as
        # the normal method, this means we need to pass "self" in (which gets
        # stripped off).

        # FIXME(morganfainberg): Does this cache actually need to be
        # invalidated? We maintain a cached revocation list, which should be
        # consulted before accepting a token as valid.  For now we will
        # do the explicit individual token invalidation.

        self._validate_token.invalidate(self, token.id)
        token_values = self.revoke_api.model.build_token_values(token)
        self.check_revocation_v3.invalidate(self, token_values)

    def revoke_token(self, token_id, revoke_chain=False):
        token = self.validate_token(token_id)

        project_id = token.project_id if token.project_scoped else None
        domain_id = token.domain_id if token.domain_scoped else None

        if revoke_chain:
            PROVIDERS.revoke_api.revoke_by_audit_chain_id(
                token.parent_audit_id,
                project_id=project_id,
                domain_id=domain_id,
            )
        else:
            PROVIDERS.revoke_api.revoke_by_audit_id(token.audit_id)

        # FIXME(morganfainberg): Does this cache actually need to be
        # invalidated? We maintain a cached revocation list, which should be
        # consulted before accepting a token as valid.  For now we will
        # do the explicit individual token invalidation.
        self.invalidate_individual_token_cache(token)



# Copyright 2017 SUSE Linux Gmbh
# Copyright 2017 Huawei
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

from oslo_db import exception as db_exception
import sqlalchemy
from sqlalchemy.ext.hybrid import hybrid_property

from keystone.common import driver_hints
from keystone.common import sql
from keystone import exception
from keystone.i18n import _
from keystone.limit.backends import base


class RegisteredLimitModel(sql.ModelBase, sql.ModelDictMixin):
    __tablename__ = 'registered_limit'
    attributes = [
        'internal_id',
        'id',
        'service_id',
        'region_id',
        'resource_name',
        'default_limit',
        'description',
    ]

    internal_id = sql.Column(sql.Integer, primary_key=True, nullable=False)
    id = sql.Column(sql.String(length=64), nullable=False, unique=True)
    service_id = sql.Column(sql.String(255), sql.ForeignKey('service.id'))
    region_id = sql.Column(
        sql.String(64), sql.ForeignKey('region.id'), nullable=True
    )
    resource_name = sql.Column(sql.String(255))
    default_limit = sql.Column(sql.Integer, nullable=False)
    description = sql.Column(sql.Text())

    def to_dict(self):
        ref = super().to_dict()
        ref.pop('internal_id')
        return ref


class LimitModel(sql.ModelBase, sql.ModelDictMixin):
    __tablename__ = 'limit'
    attributes = [
        'internal_id',
        'id',
        'project_id',
        'domain_id',
        'service_id',
        'region_id',
        'resource_name',
        'resource_limit',
        'description',
        'registered_limit_id',
    ]

    internal_id = sql.Column(sql.Integer, primary_key=True, nullable=False)
    id = sql.Column(sql.String(length=64), nullable=False, unique=True)
    project_id = sql.Column(sql.String(64))
    domain_id = sql.Column(sql.String(64))
    resource_limit = sql.Column(sql.Integer, nullable=False)
    description = sql.Column(sql.Text())
    registered_limit_id = sql.Column(
        sql.String(64), sql.ForeignKey('registered_limit.id')
    )

    registered_limit = sqlalchemy.orm.relationship('RegisteredLimitModel')

    @hybrid_property
    def service_id(self):
        if self.registered_limit:
            return self.registered_limit.service_id
        return None

    @service_id.expression  # type: ignore[no-redef]
    def service_id(self):
        return RegisteredLimitModel.service_id

    @hybrid_property
    def region_id(self):
        if self.registered_limit:
            return self.registered_limit.region_id
        return None

    @region_id.expression  # type: ignore[no-redef]
    def region_id(self):
        return RegisteredLimitModel.region_id

    @hybrid_property
    def resource_name(self):
        if self.registered_limit:
            return self.registered_limit.resource_name
        return self._resource_name

    @resource_name.expression  # type: ignore[no-redef]
    def resource_name(self):
        return RegisteredLimitModel.resource_name

    def to_dict(self):
        ref = super().to_dict()
        if self.registered_limit:
            ref['service_id'] = self.registered_limit.service_id
            ref['region_id'] = self.registered_limit.region_id
            ref['resource_name'] = self.registered_limit.resource_name
        ref.pop('internal_id')
        ref.pop('registered_limit_id')
        return ref


class UnifiedLimit(base.UnifiedLimitDriverBase):
    def _check_unified_limit_unique(
        self, unified_limit, is_registered_limit=True
    ):
        # Ensure the new created or updated unified limit won't break the
        # current reference between registered limit and limit. i.e. We should
        # ensure that there is no duplicate entry.
        hints = driver_hints.Hints()
        if is_registered_limit:
            hints.add_filter('service_id', unified_limit['service_id'])
            hints.add_filter('resource_name', unified_limit['resource_name'])
            hints.add_filter('region_id', unified_limit.get('region_id'))
            with sql.session_for_read() as session:
                query = session.query(RegisteredLimitModel)
                unified_limits = sql.filter_limit_query(
                    RegisteredLimitModel, query, hints
                ).all()
        else:
            hints.add_filter(
                'registered_limit_id', unified_limit['registered_limit_id']
            )
            is_project_limit = (
                True if unified_limit.get('project_id') else False
            )
            if is_project_limit:
                hints.add_filter('project_id', unified_limit['project_id'])
            else:
                hints.add_filter('domain_id', unified_limit['domain_id'])
            with sql.session_for_read() as session:
                query = session.query(LimitModel)
                unified_limits = sql.filter_limit_query(
                    LimitModel, query, hints
                ).all()

        if unified_limits:
            msg = _('Duplicate entry')
            limit_type = 'registered_limit' if is_registered_limit else 'limit'
            raise exception.Conflict(type=limit_type, details=msg)

    def _check_referenced_limit_reference(self, registered_limit):
        # When updating or deleting a registered limit, we should ensure there
        # is no reference limit.
        with sql.session_for_read() as session:
            limits = session.query(LimitModel).filter_by(
                registered_limit_id=registered_limit['id']
            )
        if limits.all():
            raise exception.RegisteredLimitError(id=registered_limit.id)

    @sql.handle_conflicts(conflict_type='registered_limit')
    def create_registered_limits(self, registered_limits):
        with sql.session_for_write() as session:
            new_registered_limits = []
            for registered_limit in registered_limits:
                self._check_unified_limit_unique(registered_limit)
                ref = RegisteredLimitModel.from_dict(registered_limit)
                session.add(ref)
                new_registered_limits.append(ref.to_dict())
            return new_registered_limits

    @sql.handle_conflicts(conflict_type='registered_limit')
    def update_registered_limit(self, registered_limit_id, registered_limit):
        try:
            with sql.session_for_write() as session:
                ref = self._get_registered_limit(session, registered_limit_id)
                self._check_referenced_limit_reference(ref)
                old_dict = ref.to_dict()
                old_dict.update(registered_limit)
                if (
                    registered_limit.get('service_id')
                    or 'region_id' in registered_limit
                    or registered_limit.get('resource_name')
                ):
                    self._check_unified_limit_unique(old_dict)
                new_registered_limit = RegisteredLimitModel.from_dict(old_dict)
                for attr in registered_limit:
                    if attr != 'id':
                        setattr(ref, attr, getattr(new_registered_limit, attr))
                return ref.to_dict()
        except db_exception.DBReferenceError:
            raise exception.RegisteredLimitError(id=registered_limit_id)

    @driver_hints.truncated
    def list_registered_limits(self, hints):
        with sql.session_for_read() as session:
            registered_limits = session.query(RegisteredLimitModel)
            registered_limits = sql.filter_limit_query(
                RegisteredLimitModel, registered_limits, hints
            )
            return [s.to_dict() for s in registered_limits]

    def _get_registered_limit(self, session, registered_limit_id):
        query = session.query(RegisteredLimitModel).filter_by(
            id=registered_limit_id
        )
        ref = query.first()
        if ref is None:
            raise exception.RegisteredLimitNotFound(id=registered_limit_id)
        return ref

    def get_registered_limit(self, registered_limit_id):
        with sql.session_for_read() as session:
            return self._get_registered_limit(
                session, registered_limit_id
            ).to_dict()

    def delete_registered_limit(self, registered_limit_id):
        try:
            with sql.session_for_write() as session:
                ref = self._get_registered_limit(session, registered_limit_id)
                self._check_referenced_limit_reference(ref)
                session.delete(ref)
        except db_exception.DBReferenceError:
            raise exception.RegisteredLimitError(id=registered_limit_id)

    def _check_and_fill_registered_limit_id(self, limit):
        # Make sure there is a referenced registered limit first. Then add
        # the registered limit id to the new created limit.
        hints = driver_hints.Hints()
        limit_copy = copy.deepcopy(limit)
        hints.add_filter('service_id', limit_copy.pop('service_id'))
        hints.add_filter('resource_name', limit_copy.pop('resource_name'))
        hints.add_filter('region_id', limit_copy.pop('region_id', None))

        with sql.session_for_read() as session:
            registered_limits = session.query(RegisteredLimitModel)
            registered_limits = sql.filter_limit_query(
                RegisteredLimitModel, registered_limits, hints
            )
        reg_limits = registered_limits.all()
        if not reg_limits:
            raise exception.NoLimitReference

        limit_copy['registered_limit_id'] = reg_limits[0]['id']
        return limit_copy

    @sql.handle_conflicts(conflict_type='limit')
    def create_limits(self, limits):
        try:
            with sql.session_for_write() as session:
                new_limits = []
                for limit in limits:
                    target = self._check_and_fill_registered_limit_id(limit)
                    self._check_unified_limit_unique(
                        target, is_registered_limit=False
                    )
                    ref = LimitModel.from_dict(target)
                    session.add(ref)
                    new_limit = ref.to_dict()
                    new_limit['service_id'] = limit['service_id']
                    new_limit['region_id'] = limit.get('region_id')
                    new_limit['resource_name'] = limit['resource_name']
                    new_limits.append(new_limit)
                return new_limits
        except db_exception.DBReferenceError:
            raise exception.NoLimitReference()

    @sql.handle_conflicts(conflict_type='limit')
    def update_limit(self, limit_id, limit):
        with sql.session_for_write() as session:
            ref = self._get_limit(session, limit_id)
            if limit.get('resource_limit'):
                ref.resource_limit = limit['resource_limit']
            if limit.get('description'):
                ref.description = limit['description']
            return ref.to_dict()

    @driver_hints.truncated
    def list_limits(self, hints):
        with sql.session_for_read() as session:
            query = session.query(LimitModel).outerjoin(RegisteredLimitModel)
            limits = sql.filter_limit_query(LimitModel, query, hints)
            return [limit.to_dict() for limit in limits]

    def _get_limit(self, session, limit_id):
        query = session.query(LimitModel).filter_by(id=limit_id)
        ref = query.first()
        if ref is None:
            raise exception.LimitNotFound(id=limit_id)
        return ref

    def get_limit(self, limit_id):
        with sql.session_for_read() as session:
            return self._get_limit(session, limit_id).to_dict()

    def delete_limit(self, limit_id):
        with sql.session_for_write() as session:
            ref = self._get_limit(session, limit_id)
            session.delete(ref)

    def delete_limits_for_project(self, project_id):
        limit_ids = []
        with sql.session_for_write() as session:
            query = session.query(LimitModel)
            query = query.filter_by(project_id=project_id)
            for limit in query.all():
                limit_ids.append(limit.id)
            query.delete()
        return limit_ids



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import os

from oslo_log import log
import stevedore

try:
    # werkzeug 0.15.x
    from werkzeug.middleware import proxy_fix
except ImportError:
    # werkzeug 0.14.x
    from werkzeug.contrib import fixers as proxy_fix

from keystone.common import profiler
import keystone.conf
import keystone.server
from keystone.server.flask import application
from keystone.server.flask.request_processing.middleware import auth_context
from keystone.server.flask.request_processing.middleware import url_normalize

# NOTE(morgan): Middleware Named Tuple with the following values:
#   * "namespace": namespace for the entry_point
#   * "ep": the entry-point name
#   * "conf": extra config data for the entry_point (None or Dict)
_Middleware = collections.namedtuple(
    '_Middleware', ['namespace', 'ep', 'conf']
)


CONF = keystone.conf.CONF
# NOTE(morgan): ORDER HERE IS IMPORTANT! The middleware will process the
# request in this list's order.
_APP_MIDDLEWARE = (
    _Middleware(
        namespace='keystone.server_middleware',
        ep='cors',
        conf={'oslo_config_project': 'keystone'},
    ),
    _Middleware(
        namespace='keystone.server_middleware', ep='sizelimit', conf={}
    ),
    _Middleware(
        namespace='keystone.server_middleware',
        ep='http_proxy_to_wsgi',
        conf={},
    ),
    _Middleware(
        namespace='keystone.server_middleware', ep='osprofiler', conf={}
    ),
    _Middleware(
        namespace='keystone.server_middleware', ep='request_id', conf={}
    ),
)

# NOTE(morgan): ORDER HERE IS IMPORTANT! Each of these middlewares are
# implemented/defined explicitly in Keystone Server. They do some level of
# lifting to ensure the request is properly handled. It is importat to note
# that these will be processed in the order of this list AND after all
# middleware defined in _APP_MIDDLEWARE. AuthContextMiddleware should always
# be the last element here as long as it is an actual Middleware.
_KEYSTONE_MIDDLEWARE = (
    url_normalize.URLNormalizingMiddleware,
    auth_context.AuthContextMiddleware,
)


def _get_config_files(env=None):
    if env is None:
        env = os.environ

    dirname = env.get('OS_KEYSTONE_CONFIG_DIR', '').strip()

    files = [
        s.strip()
        for s in env.get('OS_KEYSTONE_CONFIG_FILES', '').split(';')
        if s.strip()
    ]

    if dirname:
        if not files:
            files = ['keystone.conf']
        files = [os.path.join(dirname, fname) for fname in files]

    return files


def setup_app_middleware(app):
    # NOTE(morgan): Load the middleware, in reverse order, we wrap the app
    # explicitly; reverse order to ensure the first element in _APP_MIDDLEWARE
    # processes the request first.

    MW = _APP_MIDDLEWARE
    IMW = _KEYSTONE_MIDDLEWARE

    # Add in optional (config-based) middleware
    # NOTE(morgan): Each of these may need to be in a specific location
    # within the pipeline therefore cannot be magically appended/prepended
    if CONF.wsgi.debug_middleware:
        # Add in the Debug Middleware
        MW = (
            _Middleware(
                namespace='keystone.server_middleware', ep='debug', conf={}
            ),
        ) + _APP_MIDDLEWARE

    # Apply internal-only Middleware (e.g. AuthContextMiddleware). These
    # are below all externally loaded middleware in request processing.
    for mw in reversed(IMW):
        app.wsgi_app = mw(app.wsgi_app)

    # Apply the middleware to the application.
    for mw in reversed(MW):
        # TODO(morgan): Explore moving this to ExtensionManager, but we
        # want to be super careful about what middleware we load and in
        # what order. DriverManager gives us that capability and only loads
        # the entry points we care about rather than all of them.

        # Load via Stevedore, initialize the class via the factory so we can
        # initialize the "loaded" entrypoint with the currently bound
        # object pointed at "application". We may need to eventually move away
        # from the "factory" mechanism.
        loaded = stevedore.DriverManager(
            mw.namespace, mw.ep, invoke_on_load=False
        )
        # NOTE(morgan): global_conf (args[0]) to the factory is always empty
        # and local_conf (args[1]) will be the mw.conf dict. This allows for
        # configuration to be passed for middleware such as oslo CORS which
        # expects oslo_config_project or "allowed_origin" to be in the
        # local_conf, this is all a hold-over from paste-ini and pending
        # reworking/removal(s)
        factory_func = loaded.driver.factory({}, **mw.conf)
        app.wsgi_app = factory_func(app.wsgi_app)

    # Apply werkzeug specific middleware
    app.wsgi_app = proxy_fix.ProxyFix(app.wsgi_app)
    return app


def initialize_application(
    name, post_log_configured_function=lambda: None, config_files=None
):
    possible_topdir = os.path.normpath(
        os.path.join(
            os.path.abspath(__file__),
            os.pardir,
            os.pardir,
            os.pardir,
            os.pardir,
        )
    )

    dev_conf = os.path.join(possible_topdir, 'etc', 'keystone.conf')
    if not config_files:
        config_files = None
        if os.path.exists(dev_conf):
            config_files = [dev_conf]

    keystone.server.configure(config_files=config_files)

    # Log the options used when starting if we're in debug mode...
    if CONF.debug:
        CONF.log_opt_values(log.getLogger(CONF.prog), log.DEBUG)

    post_log_configured_function()

    # TODO(morgan): Provide a better mechanism than "loadapp", this was for
    # paste-deploy specific mechanisms.
    def loadapp():
        app = application.application_factory(name)
        return app

    _unused, app = keystone.server.setup_backends(
        startup_application_fn=loadapp
    )

    # setup OSprofiler notifier and enable the profiling if that is configured
    # in Keystone configuration file.
    profiler.setup(name)

    return setup_app_middleware(app)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import versionutils
from oslo_policy import policy

from keystone.common.policies import base

DEPRECATED_REASON = """
As of the Train release, the project endpoint API now understands default
roles and system-scoped tokens, making the API more granular by default without
compromising security. The new policy defaults account for these changes
automatically. Be sure to take these new defaults into consideration if you are
relying on overrides in your deployment for the project endpoint API.
"""

deprecated_list_projects_for_endpoint = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_projects_for_endpoint',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_add_endpoint_to_project = policy.DeprecatedRule(
    name=base.IDENTITY % 'add_endpoint_to_project',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_check_endpoint_in_project = policy.DeprecatedRule(
    name=base.IDENTITY % 'check_endpoint_in_project',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_list_endpoints_for_project = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_endpoints_for_project',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_remove_endpoint_from_project = policy.DeprecatedRule(
    name=base.IDENTITY % 'remove_endpoint_from_project',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)


project_endpoint_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_projects_for_endpoint',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='List projects allowed to access an endpoint.',
        operations=[
            {
                'path': ('/v3/OS-EP-FILTER/endpoints/{endpoint_id}/projects'),
                'method': 'GET',
            }
        ],
        deprecated_rule=deprecated_list_projects_for_endpoint,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'add_endpoint_to_project',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Allow project to access an endpoint.',
        operations=[
            {
                'path': (
                    '/v3/OS-EP-FILTER/projects/{project_id}/'
                    'endpoints/{endpoint_id}'
                ),
                'method': 'PUT',
            }
        ],
        deprecated_rule=deprecated_add_endpoint_to_project,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'check_endpoint_in_project',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='Check if a project is allowed to access an endpoint.',
        operations=[
            {
                'path': (
                    '/v3/OS-EP-FILTER/projects/{project_id}/'
                    'endpoints/{endpoint_id}'
                ),
                'method': 'GET',
            },
            {
                'path': (
                    '/v3/OS-EP-FILTER/projects/{project_id}/'
                    'endpoints/{endpoint_id}'
                ),
                'method': 'HEAD',
            },
        ],
        deprecated_rule=deprecated_check_endpoint_in_project,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_endpoints_for_project',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='List the endpoints a project is allowed to access.',
        operations=[
            {
                'path': ('/v3/OS-EP-FILTER/projects/{project_id}/endpoints'),
                'method': 'GET',
            }
        ],
        deprecated_rule=deprecated_list_endpoints_for_project,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'remove_endpoint_from_project',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description=(
            'Remove access to an endpoint from a project that has '
            'previously been given explicit access.'
        ),
        operations=[
            {
                'path': (
                    '/v3/OS-EP-FILTER/projects/{project_id}/'
                    'endpoints/{endpoint_id}'
                ),
                'method': 'DELETE',
            }
        ],
        deprecated_rule=deprecated_remove_endpoint_from_project,
    ),
]


def list_rules():
    return project_endpoint_policies



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import hmac

import bcrypt

from keystone.common import password_hashers
from keystone import exception


class Bcrypt(password_hashers.PasswordHasher):
    """passlib transition class for implementing bcrypt password hashing"""

    name: str = "bcrypt"
    ident_values: set[str] = {"$2$", "$2a$", "$2b$", "$2x$", "$2y$"}

    @staticmethod
    def hash(password: bytes, rounds: int = 12, **kwargs) -> str:
        """Generate password hash string with ident and params

        https://pypi.org/project/bcrypt/

        :param bytes password: Password to be hashed.
        :param int round: Count of rounds.

        :returns: String in format `$2b${rounds}${salt}{digest}`
        """
        salt: bytes = bcrypt.gensalt(rounds)
        digest: bytes = bcrypt.hashpw(password, salt)
        return digest.decode("ascii")

    @staticmethod
    def verify(password: bytes, hashed: str) -> bool:
        """Verify hashing password would be equal to the `hashed` value

        :param bytes password: Password to verify
        :param string hashed: Hashed password. Used to extract hashing
            parameters

        :returns: boolean whether hashing password with the same parameters
            would match hashed value
        """
        return bcrypt.checkpw(password, hashed.encode("ascii"))


class Bcrypt_sha256(password_hashers.PasswordHasher):
    """passlib transition class for bcrypt_sha256 password hashing"""

    name: str = "bcrypt_sha256"
    ident_values: set[str] = {"$2a$", "$2b$"}
    prefix: str = "$bcrypt-sha256$"

    @staticmethod
    def hash(password: bytes, rounds: int = 12, **kwargs) -> str:
        """Generate password hash string with ident and params

        https://pypi.org/project/bcrypt/

        :param bytes password: Password to be hashed.
        :param int round: Count of rounds.

        :returns: String in format
            `$bcrypt-sha256$r={rounds},t={ident},v={version}${salt}${digest}`
        """
        # generate salt with ident and options
        salt_with_opts: bytes = bcrypt.gensalt(rounds)
        # get the pure salt
        salt: bytes = salt_with_opts[-22:]
        # make a `str` salt
        salt_str: str = salt.decode("ascii")

        # NOTE(gtema): passlib calculates sha256 hmac digest of the password
        # with the key set to salt
        # Calculate password hmac digest with salt as key
        hmac_digest: bytes = base64.b64encode(
            hmac.digest(salt, password, "sha256")
        )

        # calculate bcrypt hash
        hashed: str = bcrypt.hashpw(hmac_digest, salt_with_opts).decode(
            "ascii"
        )
        # get the digest part of the hash
        digest: str = hashed[-31:]

        # Construct `passlib` compatible format of the bcrypt-sha256 hash
        return f"{Bcrypt_sha256.prefix}v=2,t=2b,r={rounds}${salt_str}${digest}"

    @staticmethod
    def verify(password: bytes, hashed: str) -> bool:
        """Verify hashing password would be equal to the `hashed` value

        :param bytes password: Password to verify
        :param string hashed: Hashed password. Used to extract hashing
            parameters

        :returns: boolean whether hashing password with the same parameters
            would match hashed value
        """
        opts: dict = {}
        data: str = hashed
        # Strip the ident from the hashed value
        if hashed.startswith(Bcrypt_sha256.prefix):
            data = hashed[len(Bcrypt_sha256.prefix) :]
        # split hashed string to extract parameters
        parts: list[str] = data.split("$")
        salt: str
        digest: str

        if len(parts) == 3:
            params, salt, digest = parts
        else:
            raise exception.PasswordValidationError("malformed password hash")

        for param in params.split(","):
            if param.startswith("r="):
                # Extract rounds passlib applied
                opts["r"] = int(param[2:])
            if param.startswith("t="):
                # indent applied during hashing
                opts["t"] = param[2:]

        # Calculate password hmac digest with salt as key
        hmac_digest: bytes = base64.b64encode(
            hmac.digest(salt.encode("ascii"), password, "sha256")
        )

        # Normalize salt to whatever bcrypt expects it to be
        new_salt: str = f"${opts['t']}${opts['r']}${salt}"

        # verify_digest: str = bcrypt.hashpw(
        #     hmac_digest.encode("ascii"), new_salt.encode("ascii")
        # )[-31:].decode("ascii")

        # Invoke bcrypt checkpw with the re-calculated salt
        return bcrypt.checkpw(
            hmac_digest, f"{new_salt}{digest}".encode("ascii")
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
from typing import Any

from keystone.api.validation import parameter_types
from keystone.api.validation import response_types
from keystone.common import validation
import keystone.conf
from keystone.identity.backends import resource_options as ro

CONF = keystone.conf.CONF


_identity_name: dict[str, Any] = {
    'type': 'string',
    'minLength': 1,
    'maxLength': 255,
    'pattern': r'[\S]+',
}

# Schema for Identity v3 API

user_index_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {
        "domain_id": parameter_types.domain_id,
        "enabled": {
            "description": "Whether the identity provider is enabled or not",
            **parameter_types.boolean,
        },
        "idp_id": {
            "type": "string",
            "description": "Filters the response by an identity provider ID.",
        },
        "name": parameter_types.name,
        "password_expires_at": {
            "type": "string",
            "description": (
                "Filter results based on which user passwords have "
                "expired. The query should include an operator and "
                "a timestamp with a colon (:) separating the two, for "
                "example: `password_expires_at={operator}:{timestamp}`\n"
                "Valid operators are: lt, lte, gt, gte, eq, and neq\n"
                "  - lt: expiration time lower than the timestamp\n"
                "  - lte: expiration time lower than or equal to the timestamp\n"
                "  - gt: expiration time higher than the timestamp\n"
                "  - gte: expiration time higher than or equal to the timestamp\n"
                "  - eq: expiration time equal to the timestamp\n"
                "  - neq: expiration time not equal to the timestamp\n\n"
                "Valid timestamps are of the form: `YYYY-MM-DDTHH:mm:ssZ`."
                "For example:"
                "`/v3/users?password_expires_at=lt:2016-12-08T22:02:00Z`\n"
                "The example would return a list of users whose password "
                "expired before the timestamp `(2016-12-08T22:02:00Z).`"
            ),
        },
        "protocol_id": {
            "type": "string",
            "description": "Filters the response by a protocol ID.",
        },
        "unique_id": {
            "type": "string",
            "description": "Filters the response by a unique ID.",
        },
        "marker": {
            "type": "string",
            "description": "ID of the last fetched entry",
        },
        "limit": {"type": ["integer", "string"]},
        "sort_key": parameter_types.sort_key,
        "sort_dir": parameter_types.sort_dir,
    },
    "additionalProperties": True,
}

_user_properties: dict[str, Any] = {
    'id': {"type": "string", "description": "The user ID.", "readOnly": True},
    'default_project_id': validation.nullable(parameter_types.id_string),
    'description': validation.nullable(parameter_types.description),
    'domain_id': parameter_types.id_string,
    'enabled': parameter_types.boolean,
    'federated': {
        'type': 'array',
        'items': {
            'type': 'object',
            'properties': {
                'idp_id': {'type': 'string'},
                'protocols': {
                    'type': 'array',
                    'items': {
                        'type': 'object',
                        'properties': {
                            'protocol_id': {'type': 'string'},
                            'unique_id': {'type': 'string'},
                        },
                        'required': ['protocol_id', 'unique_id'],
                    },
                    'minItems': 1,
                },
            },
            'required': ['idp_id', 'protocols'],
        },
    },
    'links': response_types.links,
    'name': _identity_name,
    'password_expires_at': {
        "type": ["string", "null"],
        "format": "date-time",
        "description": (
            "The date and time when the password expires. The time zone is UTC. "
            "This is a response object attribute; not valid for requests. A "
            "null value indicates that the password never expires."
        ),
        "readOnly": True,
    },
    'options': ro.USER_OPTIONS_REGISTRY.json_schema,
}

user_schema: dict[str, Any] = {
    "type": "object",
    "properties": _user_properties,
    # NOTE(gtema) User resource supports additional attributes which are stored
    # in the `extra` DB field
    "additionalProperties": True,
}

user_index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "users": {"type": "array", "items": user_schema},
        "links": response_types.links,
        "truncated": response_types.truncated,
    },
    "additionalProperties": False,
}

user_get_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {"user": user_schema},
    "required": ["user"],
    "additionalProperties": False,
}

user_create_request: dict[str, Any] = {
    "type": "object",
    "properties": {
        "user": {
            "type": "object",
            "properties": {
                "password": {"type": ["string", "null"]},
                **_user_properties,
            },
            "required": ["name"],
            "additionalProperties": True,
        }
    },
    "required": ["user"],
    "additionalProperties": False,
}

user_create_response_body: dict[str, Any] = user_get_response_body

user_update_properties = copy.deepcopy(_user_properties)
# It is not allowed anymore to update domain of the existing user
user_update_properties.pop("domain_id", None)
user_update_request: dict[str, Any] = {
    'type': 'object',
    'properties': {
        'user': {
            'type': 'object',
            'properties': {
                'password': {'type': ['string', 'null']},
                **user_update_properties,
            },
            'minProperties': 1,
            'additionalProperties': True,
        }
    },
    'required': ['user'],
    'additionalProperties': False,
}

user_update_response_body: dict[str, Any] = user_get_response_body

group_index_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {
        "domain_id": parameter_types.domain_id,
        "name": parameter_types.name,
        "marker": {
            "type": "string",
            "description": "ID of the last fetched entry",
        },
        "limit": {"type": ["integer", "string"]},
        "sort_key": parameter_types.sort_key,
        "sort_dir": parameter_types.sort_dir,
    },
    "additionalProperties": False,
}

_group_properties: dict[str, Any] = {
    'description': validation.nullable(parameter_types.description),
    'domain_id': parameter_types.id_string,
    'id': {"type": "string", "description": "The user ID.", "readOnly": True},
    'name': _identity_name,
}

group_schema: dict[str, Any] = {
    "type": "object",
    "properties": _group_properties,
    # NOTE(gtema) Group resource supports additional attributes which are stored
    # in the `extra` DB field
    "additionalProperties": True,
}

group_index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "groups": {
            "type": "array",
            "items": group_schema,
            "description": "A list of group objects",
        },
        "links": response_types.links,
        "truncated": response_types.truncated,
    },
    "required": ["groups"],
    "additionalProperties": False,
}

group_get_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {"group": group_schema},
    "required": ["group"],
    "additionalProperties": False,
}

group_create_request_body: dict[str, Any] = {
    'type': 'object',
    'properties': {
        'group': {
            'type': 'object',
            'properties': _group_properties,
            'required': ['name'],
            'additionalProperties': True,
        }
    },
    "required": ["group"],
    "additionalProperties": False,
}

group_create_response_body = group_get_response_body

group_update_properties = copy.deepcopy(_group_properties)
# It is not allowed anymore to update domain of the existing group
group_update_properties.pop("domain_id", None)
group_update_request_body: dict[str, Any] = {
    'type': 'object',
    'properties': {
        'group': {
            'type': 'object',
            'properties': group_update_properties,
            'minProperties': 1,
            'additionalProperties': True,
        }
    },
    "required": ["group"],
    'additionalProperties': False,
}

group_update_response_body = group_get_response_body

_password_change_properties = {
    'original_password': {'type': 'string'},
    'password': {'type': 'string'},
}
if getattr(CONF, 'strict_password_check', None):
    _password_change_properties['password']['maxLength'] = (
        CONF.identity.max_password_length
    )

if getattr(CONF, 'security_compliance', None):
    if getattr(CONF.security_compliance, 'password_regex', None):
        _password_change_properties['password']['pattern'] = (
            CONF.security_compliance.password_regex
        )

password_change = {
    'type': 'object',
    'properties': _password_change_properties,
    'required': ['original_password', 'password'],
    'additionalProperties': False,
}



# Copyright 2013 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from oslo_config import generator

import keystone.conf
from keystone.tests import unit

CONF = keystone.conf.CONF


class ConfigTestCase(unit.TestCase):
    def config_files(self):
        config_files = super().config_files()

        sample_file = 'keystone.conf.sample'
        args = [
            '--namespace',
            'keystone',
            '--output-file',
            unit.dirs.etc(sample_file),
        ]

        generator.main(args=args)
        config_files.insert(0, unit.dirs.etc(sample_file))
        self.addCleanup(os.remove, unit.dirs.etc(sample_file))
        return config_files

    def test_config_default(self):
        self.assertIsNone(CONF.auth.password)
        self.assertIsNone(CONF.auth.token)
        # Check config.set_config_defaults() has set [profiler]enabled.
        self.assertEqual(False, CONF.profiler.enabled)



# Copyright 2012 OpenStack Foundation
# Copyright 2013 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import fixtures
import ldappool

from keystone.common import provider_api
import keystone.conf
from keystone.identity.backends import ldap
from keystone.identity.backends.ldap import common as common_ldap
from keystone.tests import unit
from keystone.tests.unit import fakeldap
from keystone.tests.unit import test_backend_ldap

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class LdapPoolCommonTestMixin:
    """LDAP pool specific common tests used here and in live tests."""

    def cleanup_pools(self):
        common_ldap.PooledLDAPHandler.connection_pools.clear()

    def test_handler_with_use_pool_enabled(self):
        # by default use_pool and use_auth_pool is enabled in test pool config
        user_ref = PROVIDERS.identity_api.get_user(self.user_foo['id'])
        self.user_foo.pop('password')
        self.assertDictEqual(self.user_foo, user_ref)

        handler = common_ldap._get_connection(CONF.ldap.url, use_pool=True)
        self.assertIsInstance(handler, common_ldap.PooledLDAPHandler)

    @mock.patch.object(common_ldap.KeystoneLDAPHandler, 'connect')
    @mock.patch.object(common_ldap.KeystoneLDAPHandler, 'simple_bind_s')
    def test_handler_with_use_pool_not_enabled(
        self, bind_method, connect_method
    ):
        self.config_fixture.config(group='ldap', use_pool=False)
        self.config_fixture.config(group='ldap', use_auth_pool=True)
        self.cleanup_pools()

        user_api = ldap.UserApi(CONF)
        handler = user_api.get_connection(
            user=None, password=None, end_user_auth=True
        )
        # use_auth_pool flag does not matter when use_pool is False
        # still handler is non pool version
        self.assertIsInstance(handler.conn, common_ldap.PythonLDAPHandler)

    @mock.patch.object(common_ldap.KeystoneLDAPHandler, 'connect')
    @mock.patch.object(common_ldap.KeystoneLDAPHandler, 'simple_bind_s')
    def test_handler_with_end_user_auth_use_pool_not_enabled(
        self, bind_method, connect_method
    ):
        # by default use_pool is enabled in test pool config
        # now disabling use_auth_pool flag to test handler instance
        self.config_fixture.config(group='ldap', use_auth_pool=False)
        self.cleanup_pools()

        user_api = ldap.UserApi(CONF)
        handler = user_api.get_connection(
            user=None, password=None, end_user_auth=True
        )
        self.assertIsInstance(handler.conn, common_ldap.PythonLDAPHandler)

        # For end_user_auth case, flag should not be false otherwise
        # it will use, admin connections ldap pool
        handler = user_api.get_connection(
            user=None, password=None, end_user_auth=False
        )
        self.assertIsInstance(handler.conn, common_ldap.PooledLDAPHandler)

    def test_pool_size_set(self):
        # get related connection manager instance
        ldappool_cm = self.conn_pools[CONF.ldap.url]
        self.assertEqual(CONF.ldap.pool_size, ldappool_cm.size)

    def test_pool_retry_max_set(self):
        # get related connection manager instance
        ldappool_cm = self.conn_pools[CONF.ldap.url]
        self.assertEqual(CONF.ldap.pool_retry_max, ldappool_cm.retry_max)

    def test_pool_retry_delay_set(self):
        # just make one identity call to initiate ldap connection if not there
        PROVIDERS.identity_api.get_user(self.user_foo['id'])

        # get related connection manager instance
        ldappool_cm = self.conn_pools[CONF.ldap.url]
        self.assertEqual(CONF.ldap.pool_retry_delay, ldappool_cm.retry_delay)

    def test_pool_use_tls_set(self):
        # get related connection manager instance
        ldappool_cm = self.conn_pools[CONF.ldap.url]
        self.assertEqual(CONF.ldap.use_tls, ldappool_cm.use_tls)

    def test_pool_timeout_set(self):
        # get related connection manager instance
        ldappool_cm = self.conn_pools[CONF.ldap.url]
        self.assertEqual(
            CONF.ldap.pool_connection_timeout, ldappool_cm.timeout
        )

    def test_pool_use_pool_set(self):
        # get related connection manager instance
        ldappool_cm = self.conn_pools[CONF.ldap.url]
        self.assertEqual(CONF.ldap.use_pool, ldappool_cm.use_pool)

    def test_pool_connection_lifetime_set(self):
        # get related connection manager instance
        ldappool_cm = self.conn_pools[CONF.ldap.url]
        self.assertEqual(
            CONF.ldap.pool_connection_lifetime, ldappool_cm.max_lifetime
        )

    def test_max_connection_error_raised(self):
        who = CONF.ldap.user
        cred = CONF.ldap.password
        # get related connection manager instance
        ldappool_cm = self.conn_pools[CONF.ldap.url]
        ldappool_cm.size = 2

        # 3rd connection attempt should raise Max connection error
        with ldappool_cm.connection(who, cred) as _:  # conn1
            with ldappool_cm.connection(who, cred) as _:  # conn2
                try:
                    with ldappool_cm.connection(who, cred) as _:  # conn3
                        _.unbind_s()
                        self.fail()
                except Exception as ex:
                    self.assertIsInstance(
                        ex, ldappool.MaxConnectionReachedError
                    )
        ldappool_cm.size = CONF.ldap.pool_size

    def test_pool_size_expands_correctly(self):
        who = CONF.ldap.user
        cred = CONF.ldap.password
        # get related connection manager instance
        ldappool_cm = self.conn_pools[CONF.ldap.url]
        ldappool_cm.size = 3

        def _get_conn():
            return ldappool_cm.connection(who, cred)

        # Open 3 connections first
        with _get_conn() as _:  # conn1
            self.assertEqual(1, len(ldappool_cm))
            with _get_conn() as _:  # conn2
                self.assertEqual(2, len(ldappool_cm))
                with _get_conn() as _:  # conn2
                    _.unbind_ext_s()
                    self.assertEqual(3, len(ldappool_cm))

        # Then open 3 connections again and make sure size does not grow
        # over 3
        with _get_conn() as c1:  # conn1
            self.assertEqual(3, len(ldappool_cm))
            c1.connected = False
            with _get_conn() as c2:  # conn2
                self.assertEqual(3, len(ldappool_cm))
                c2.connected = False
                with _get_conn() as c3:  # conn3
                    c3.connected = False
                    c3.unbind_ext_s()
                    self.assertEqual(3, len(ldappool_cm))

        with _get_conn() as c1:  # conn1
            self.assertEqual(1, len(ldappool_cm))
            with _get_conn() as c2:  # conn2
                self.assertEqual(2, len(ldappool_cm))
                with _get_conn() as c3:  # conn3
                    c3.unbind_ext_s()
                    self.assertEqual(3, len(ldappool_cm))

    def test_password_change_with_pool(self):
        old_password = self.user_sna['password']
        self.cleanup_pools()

        # authenticate so that connection is added to pool before password
        # change
        with self.make_request():
            user_ref = PROVIDERS.identity_api.authenticate(
                user_id=self.user_sna['id'], password=self.user_sna['password']
            )

        self.user_sna.pop('password')
        self.user_sna['enabled'] = True
        self.assertUserDictEqual(self.user_sna, user_ref)

        new_password = 'new_password'
        user_ref['password'] = new_password
        PROVIDERS.identity_api.update_user(user_ref['id'], user_ref)

        # now authenticate again to make sure new password works with
        # connection pool
        with self.make_request():
            user_ref2 = PROVIDERS.identity_api.authenticate(
                user_id=self.user_sna['id'], password=new_password
            )

        user_ref.pop('password')
        self.assertUserDictEqual(user_ref, user_ref2)

        # Authentication with old password would not work here as there
        # is only one connection in pool which get bind again with updated
        # password..so no old bind is maintained in this case.
        with self.make_request():
            self.assertRaises(
                AssertionError,
                PROVIDERS.identity_api.authenticate,
                user_id=self.user_sna['id'],
                password=old_password,
            )

    @mock.patch.object(fakeldap.FakeLdap, 'search_ext')
    def test_search_ext_ensure_pool_connection_released(self, mock_search_ext):
        """Test search_ext exception resiliency.

        Call search_ext function in isolation. Doing so will cause
        search_ext to borrow a connection from the pool and associate
        it with an AsynchronousMessage object. Borrowed connection ought
        to be released if anything goes wrong during LDAP API call. This
        test case intentionally throws an exception to ensure everything
        goes as expected when LDAP connection raises an exception.
        """

        class CustomDummyException(Exception):
            pass

        # Throw an exception intentionally when LDAP
        # connection search_ext function is called
        mock_search_ext.side_effect = CustomDummyException()
        self.config_fixture.config(group='ldap', pool_size=1)
        pool = self.conn_pools[CONF.ldap.url]
        user_api = ldap.UserApi(CONF)

        # setUp primes the pool so pool
        # must have one connection
        self.assertEqual(1, len(pool))
        for i in range(1, 10):
            handler = user_api.get_connection()
            # Just to ensure that we're using pooled connections
            self.assertIsInstance(handler.conn, common_ldap.PooledLDAPHandler)
            # LDAP API will throw CustomDummyException. In this scenario
            # we expect LDAP connection to be made available back to the
            # pool.
            self.assertRaises(
                CustomDummyException,
                lambda: handler.search_ext(
                    'dc=example,dc=test',
                    'dummy',
                    'objectclass=*',
                    ['mail', 'userPassword'],
                ),
            )
            # Pooled connection must not be evicted from the pool
            self.assertEqual(1, len(pool))
            # Ensure that the connection is inactive afterwards
            with pool._pool_lock:
                for slot, conn in enumerate(pool._pool):
                    self.assertFalse(conn.active)

            self.assertEqual(mock_search_ext.call_count, i)

    @mock.patch.object(fakeldap.FakeLdap, 'result3')
    def test_result3_ensure_pool_connection_released(self, mock_result3):
        """Test search_ext-->result3 exception resiliency.

        Call search_ext function, grab an AsynchronousMessage object and
        call result3 with it. During the result3 call, LDAP API will throw
        an exception.The expectation is that the associated LDAP pool
        connection for AsynchronousMessage must be released back to the
        LDAP connection pool.
        """

        class CustomDummyException(Exception):
            pass

        # Throw an exception intentionally when LDAP
        # connection result3 function is called
        mock_result3.side_effect = CustomDummyException()
        self.config_fixture.config(group='ldap', pool_size=1)
        pool = self.conn_pools[CONF.ldap.url]
        user_api = ldap.UserApi(CONF)

        # setUp primes the pool so pool
        # must have one connection
        self.assertEqual(1, len(pool))
        for i in range(1, 10):
            handler = user_api.get_connection()
            # Just to ensure that we're using pooled connections
            self.assertIsInstance(handler.conn, common_ldap.PooledLDAPHandler)
            msg = handler.search_ext(
                'dc=example,dc=test',
                'dummy',
                'objectclass=*',
                ['mail', 'userPassword'],
            )
            # Connection is in use, must be already marked active
            self.assertTrue(msg.connection.active)
            # Pooled connection must not be evicted from the pool
            self.assertEqual(1, len(pool))
            # LDAP API will throw CustomDummyException. In this
            # scenario we expect LDAP connection to be made
            # available back to the pool.
            self.assertRaises(
                CustomDummyException, lambda: handler.result3(msg)
            )
            # Connection must be set inactive
            self.assertFalse(msg.connection.active)
            # Pooled connection must not be evicted from the pool
            self.assertEqual(1, len(pool))
            self.assertEqual(mock_result3.call_count, i)


class LDAPIdentity(
    LdapPoolCommonTestMixin, test_backend_ldap.LDAPIdentity, unit.TestCase
):
    """Executes tests in existing base class with pooled LDAP handler."""

    def setUp(self):
        self.useFixture(
            fixtures.MockPatchObject(
                common_ldap.PooledLDAPHandler,
                'Connector',
                fakeldap.FakeLdapPool,
            )
        )
        super().setUp()

        self.addCleanup(self.cleanup_pools)
        # storing to local variable to avoid long references
        self.conn_pools = common_ldap.PooledLDAPHandler.connection_pools
        # super class loads db fixtures which establishes ldap connection
        # so adding dummy call to highlight connection pool initialization
        # as its not that obvious though its not needed here
        PROVIDERS.identity_api.get_user(self.user_foo['id'])

    def config_files(self):
        config_files = super().config_files()
        config_files.append(unit.dirs.tests_conf('backend_ldap_pool.conf'))
        return config_files



# Copyright 2012 OpenStack Foundation
# Copyright 2012 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Main entry point into the Catalog service."""

from keystone.common import cache
from keystone.common import driver_hints
from keystone.common import manager
from keystone.common import provider_api
import keystone.conf
from keystone import exception
from keystone.i18n import _
from keystone import notifications

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


# This is a general cache region for catalog administration (CRUD operations).
MEMOIZE = cache.get_memoization_decorator(group='catalog')

# This builds a discrete cache region dedicated to complete service catalogs
# computed for a given user + project pair. Any write operation to create,
# modify or delete elements of the service catalog should invalidate this
# entire cache region.
COMPUTED_CATALOG_REGION = cache.create_region(name='computed catalog region')
MEMOIZE_COMPUTED_CATALOG = cache.get_memoization_decorator(
    group='catalog', region=COMPUTED_CATALOG_REGION
)


class Manager(manager.Manager):
    """Default pivot point for the Catalog backend.

    See :mod:`keystone.common.manager.Manager` for more details on how this
    dynamically calls the backend.

    """

    driver_namespace = 'keystone.catalog'
    _provides_api = 'catalog_api'

    _ENDPOINT = 'endpoint'
    _SERVICE = 'service'
    _REGION = 'region'

    def __init__(self):
        super().__init__(CONF.catalog.driver)
        notifications.register_event_callback(
            notifications.ACTIONS.deleted,
            'project',
            self._on_project_or_endpoint_delete,
        )
        notifications.register_event_callback(
            notifications.ACTIONS.deleted,
            'endpoint',
            self._on_project_or_endpoint_delete,
        )

    def _on_project_or_endpoint_delete(
        self, service, resource_type, operation, payload
    ):
        project_or_endpoint_id = payload['resource_info']
        if resource_type == 'project':
            PROVIDERS.catalog_api.delete_association_by_project(
                project_or_endpoint_id
            )
            PROVIDERS.catalog_api.delete_endpoint_group_association_by_project(
                project_or_endpoint_id
            )
        else:
            PROVIDERS.catalog_api.delete_association_by_endpoint(
                project_or_endpoint_id
            )

    def create_region(self, region_ref, initiator=None):
        # Check duplicate ID
        try:
            self.get_region(region_ref['id'])
        except exception.RegionNotFound:  # nosec
            # A region with the same id doesn't exist already, good.
            pass
        else:
            msg = _('Duplicate ID, %s.') % region_ref['id']
            raise exception.Conflict(type='region', details=msg)

        # NOTE(lbragstad,dstanek): The description column of the region
        # database cannot be null. So if the user doesn't pass in a
        # description or passes in a null description then set it to an
        # empty string.
        if region_ref.get('description') is None:
            region_ref['description'] = ''
        try:
            ret = self.driver.create_region(region_ref)
        except exception.NotFound:
            parent_region_id = region_ref.get('parent_region_id')
            raise exception.RegionNotFound(region_id=parent_region_id)

        notifications.Audit.created(self._REGION, ret['id'], initiator)
        COMPUTED_CATALOG_REGION.invalidate()
        return ret

    @MEMOIZE
    def get_region(self, region_id):
        try:
            return self.driver.get_region(region_id)
        except exception.NotFound:
            raise exception.RegionNotFound(region_id=region_id)

    def update_region(self, region_id, region_ref, initiator=None):
        # NOTE(lbragstad,dstanek): The description column of the region
        # database cannot be null. So if the user passes in a null
        # description set it to an empty string.
        if 'description' in region_ref and region_ref['description'] is None:
            region_ref['description'] = ''
        ref = self.driver.update_region(region_id, region_ref)
        notifications.Audit.updated(self._REGION, region_id, initiator)
        self.get_region.invalidate(self, region_id)
        COMPUTED_CATALOG_REGION.invalidate()
        return ref

    def delete_region(self, region_id, initiator=None):
        try:
            ret = self.driver.delete_region(region_id)
            notifications.Audit.deleted(self._REGION, region_id, initiator)
            self.get_region.invalidate(self, region_id)
            COMPUTED_CATALOG_REGION.invalidate()
            return ret
        except exception.NotFound:
            raise exception.RegionNotFound(region_id=region_id)

    @manager.response_truncated
    def list_regions(self, hints=None):
        return self.driver.list_regions(hints or driver_hints.Hints())

    def create_service(self, service_id, service_ref, initiator=None):
        service_ref.setdefault('enabled', True)
        service_ref.setdefault('name', '')
        ref = self.driver.create_service(service_id, service_ref)
        notifications.Audit.created(self._SERVICE, service_id, initiator)
        COMPUTED_CATALOG_REGION.invalidate()
        return ref

    @MEMOIZE
    def get_service(self, service_id):
        try:
            return self.driver.get_service(service_id)
        except exception.NotFound:
            raise exception.ServiceNotFound(service_id=service_id)

    def update_service(self, service_id, service_ref, initiator=None):
        ref = self.driver.update_service(service_id, service_ref)
        notifications.Audit.updated(self._SERVICE, service_id, initiator)
        self.get_service.invalidate(self, service_id)
        COMPUTED_CATALOG_REGION.invalidate()
        return ref

    def delete_service(self, service_id, initiator=None):
        try:
            endpoints = self.list_endpoints()
            ret = self.driver.delete_service(service_id)
            notifications.Audit.deleted(self._SERVICE, service_id, initiator)
            self.get_service.invalidate(self, service_id)
            for endpoint in endpoints:
                if endpoint['service_id'] == service_id:
                    self.get_endpoint.invalidate(self, endpoint['id'])
            COMPUTED_CATALOG_REGION.invalidate()
            return ret
        except exception.NotFound:
            raise exception.ServiceNotFound(service_id=service_id)

    @manager.response_truncated
    def list_services(self, hints=None):
        return self.driver.list_services(hints or driver_hints.Hints())

    def _assert_region_exists(self, region_id):
        try:
            if region_id is not None:
                self.get_region(region_id)
        except exception.RegionNotFound:
            raise exception.ValidationError(
                attribute='endpoint region_id', target='region table'
            )

    def _assert_service_exists(self, service_id):
        try:
            if service_id is not None:
                self.get_service(service_id)
        except exception.ServiceNotFound:
            raise exception.ValidationError(
                attribute='endpoint service_id', target='service table'
            )

    def create_endpoint(self, endpoint_id, endpoint_ref, initiator=None):
        self._assert_region_exists(endpoint_ref.get('region_id'))
        self._assert_service_exists(endpoint_ref['service_id'])
        ref = self.driver.create_endpoint(endpoint_id, endpoint_ref)

        notifications.Audit.created(self._ENDPOINT, endpoint_id, initiator)
        COMPUTED_CATALOG_REGION.invalidate()
        return ref

    def update_endpoint(self, endpoint_id, endpoint_ref, initiator=None):
        self._assert_region_exists(endpoint_ref.get('region_id'))
        self._assert_service_exists(endpoint_ref.get('service_id'))
        ref = self.driver.update_endpoint(endpoint_id, endpoint_ref)
        notifications.Audit.updated(self._ENDPOINT, endpoint_id, initiator)
        self.get_endpoint.invalidate(self, endpoint_id)
        COMPUTED_CATALOG_REGION.invalidate()
        return ref

    def delete_endpoint(self, endpoint_id, initiator=None):
        try:
            ret = self.driver.delete_endpoint(endpoint_id)
            notifications.Audit.deleted(self._ENDPOINT, endpoint_id, initiator)
            self.get_endpoint.invalidate(self, endpoint_id)
            COMPUTED_CATALOG_REGION.invalidate()
            return ret
        except exception.NotFound:
            raise exception.EndpointNotFound(endpoint_id=endpoint_id)

    @MEMOIZE
    def get_endpoint(self, endpoint_id):
        try:
            return self.driver.get_endpoint(endpoint_id)
        except exception.NotFound:
            raise exception.EndpointNotFound(endpoint_id=endpoint_id)

    @manager.response_truncated
    def list_endpoints(self, hints=None):
        return self.driver.list_endpoints(hints or driver_hints.Hints())

    @MEMOIZE_COMPUTED_CATALOG
    def get_v3_catalog(self, user_id, project_id):
        return self.driver.get_v3_catalog(user_id, project_id)

    def add_endpoint_to_project(self, endpoint_id, project_id):
        self.driver.add_endpoint_to_project(endpoint_id, project_id)
        COMPUTED_CATALOG_REGION.invalidate()

    def remove_endpoint_from_project(self, endpoint_id, project_id):
        self.driver.remove_endpoint_from_project(endpoint_id, project_id)
        COMPUTED_CATALOG_REGION.invalidate()

    def add_endpoint_group_to_project(self, endpoint_group_id, project_id):
        self.driver.add_endpoint_group_to_project(
            endpoint_group_id, project_id
        )
        COMPUTED_CATALOG_REGION.invalidate()

    def remove_endpoint_group_from_project(
        self, endpoint_group_id, project_id
    ):
        self.driver.remove_endpoint_group_from_project(
            endpoint_group_id, project_id
        )
        COMPUTED_CATALOG_REGION.invalidate()

    def delete_endpoint_group_association_by_project(self, project_id):
        try:
            self.driver.delete_endpoint_group_association_by_project(
                project_id
            )
        except exception.NotImplemented:
            # Some catalog drivers don't support this
            pass

    def get_endpoint_groups_for_project(self, project_id):
        # recover the project endpoint group memberships and for each
        # membership recover the endpoint group
        PROVIDERS.resource_api.get_project(project_id)
        try:
            refs = self.list_endpoint_groups_for_project(project_id)
            endpoint_groups = [
                self.get_endpoint_group(ref['endpoint_group_id'])
                for ref in refs
            ]
            return endpoint_groups
        except exception.EndpointGroupNotFound:
            return []

    def get_endpoints_filtered_by_endpoint_group(self, endpoint_group_id):
        endpoints = self.list_endpoints()
        filters = self.get_endpoint_group(endpoint_group_id)['filters']
        filtered_endpoints = []

        for endpoint in endpoints:
            is_candidate = True
            for key, value in filters.items():
                if endpoint[key] != value:
                    is_candidate = False
                    break
            if is_candidate:
                filtered_endpoints.append(endpoint)
        return filtered_endpoints

    def list_endpoints_for_project(self, project_id):
        """List all endpoints associated with a project.

        :param project_id: project identifier to check
        :type project_id: string
        :returns: a list of endpoint ids or an empty list.

        """
        refs = self.driver.list_endpoints_for_project(project_id)
        filtered_endpoints = {}
        for ref in refs:
            try:
                endpoint = self.get_endpoint(ref['endpoint_id'])
                filtered_endpoints.update({ref['endpoint_id']: endpoint})
            except exception.EndpointNotFound:
                # remove bad reference from association
                self.remove_endpoint_from_project(
                    ref['endpoint_id'], project_id
                )

        # need to recover endpoint_groups associated with project
        # then for each endpoint group return the endpoints.
        endpoint_groups = self.get_endpoint_groups_for_project(project_id)
        for endpoint_group in endpoint_groups:
            endpoint_refs = self.get_endpoints_filtered_by_endpoint_group(
                endpoint_group['id']
            )
            # now check if any endpoints for current endpoint group are not
            # contained in the list of filtered endpoints
            for endpoint_ref in endpoint_refs:
                if endpoint_ref['id'] not in filtered_endpoints:
                    filtered_endpoints[endpoint_ref['id']] = endpoint_ref

        return filtered_endpoints

    def delete_association_by_endpoint(self, endpoint_id):
        try:
            self.driver.delete_association_by_endpoint(endpoint_id)
        except exception.NotImplemented:
            # Some catalog drivers don't support this
            pass

    def delete_association_by_project(self, project_id):
        try:
            self.driver.delete_association_by_project(project_id)
        except exception.NotImplemented:
            # Some catalog drivers don't support this
            pass
